@@ -6,7 +6,7 @@
 # of the GNU General Public License, incorporated herein by reference.
 
 import cStringIO
-import os
+import os, re
 
 from mercurial import hg, util, patch, commands, cmdutil
 from mercurial import match as matchmod, ui as uimod
@@ -88,9 +88,9 @@ class ChunksWidget(QWidget):
         for name, desc, icon, key, tip, cb in [
             ('diff', _('Visual Diff'), 'visualdiff', 'Ctrl+D',
               _('View file changes in external diff tool'), self.vdiff),
-            ('edit', _('Edit Local'), 'edit-file', 'Shift+Ctrl+E',
+            ('edit', _('Edit Local'), 'edit-file', 'Shift+Ctrl+L',
               _('Edit current file in working copy'), self.editCurrentFile),
-            ('revert', _('Revert to Revision'), 'hg-revert', 'Alt+Ctrl+T',
+            ('revert', _('Revert to Revision'), 'hg-revert', 'Shift+Ctrl+R',
               _('Revert file(s) to contents at this revision'),
               self.revertfile),
             ]:
@@ -177,17 +177,10 @@ class ChunksWidget(QWidget):
                 eolmode = 'strict'
             else:
                 eolmode = eolmode.lower()
+            # 'updatestate' flag has no effect since hg 1.9
             try:
-                # hg-1.9
                 ret = patch.internalpatch(ui, repo, fp, 1, files=None,
                                           eolmode=eolmode, similarity=0)
-            except TypeError:
-                # hg-1.8
-                pfiles = {}
-                ret = patch.internalpatch(fp, ui, 1, repo.root, pfiles,
-                                          eolmode=eolmode)
-                if updatestate:
-                    cmdutil.updatedir(repo.ui, repo, pfiles)
             except ValueError:
                 ret = -1
             if ret < 0:
@@ -196,8 +189,9 @@ class ChunksWidget(QWidget):
         except (patch.PatchError, EnvironmentError), err:
             ok = False
             self.showMessage.emit(hglib.tounicode(str(err)))
+        rejfilere = re.compile(r'\b%s\.rej\b' % re.escape(wfile))
         for line in ui.popbuffer().splitlines():
-            if line.endswith(wfile + '.rej'):
+            if rejfilere.search(line):
                 if qtlib.QuestionMsgBox(_('Manually resolve rejected chunks?'),
                                         hglib.tounicode(line) + u'<br><br>' +
                                         _('Edit patched file and rejects?'),
@@ -236,12 +230,15 @@ class ChunksWidget(QWidget):
         if not dchunks:
             self.showMessage.emit(_('No deletable chunks'))
             return
+        ctx = self.ctx
         kchunks = [c for c in chunks[1:] if not c.selected]
         revertall = False
-        if not kchunks and qtlib.QuestionMsgBox(_('No chunks remain'),
-                                                _('Remove all file changes?')):
-            revertall = True
-        ctx = self.ctx
+        if not kchunks:
+            if isinstance(ctx, patchctx):
+                revertmsg = _('Completely remove file from patch?')
+            else:
+                revertmsg = _('Revert all file changes?')
+            revertall = qtlib.QuestionMsgBox(_('No chunks remain'), revertmsg)
         if isinstance(ctx, patchctx):
             repo.thgbackup(ctx._path)
             fp = util.atomictempfile(ctx._path, 'wb')
@@ -510,6 +507,7 @@ class DiffBrowser(QFrame):
         self.countselected = 0
         self._ctx = None
         self._lastfile = None
+        self._status = None
 
         vbox = QVBoxLayout()
         vbox.setContentsMargins(0,0,0,0)
@@ -553,6 +551,7 @@ class DiffBrowser(QFrame):
         self.layout().addSpacing(2)
         w.hide()
 
+        self._forceviewindicator = None
         self.sci = qscilib.Scintilla(self)
         self.sci.setReadOnly(True)
         self.sci.setUtf8(True)
@@ -566,8 +565,15 @@ class DiffBrowser(QFrame):
         self.sci.setMarginWidth(1, QFontMetrics(self.font()).width('XX'))
         self.sci.setMarginSensitivity(1, True)
         self.sci.marginClicked.connect(self.marginClicked)
-        self.selected = self.sci.markerDefine(qsci.Plus, -1)
-        self.unselected = self.sci.markerDefine(qsci.Minus, -1)
+
+        self._checkedpix = qtlib.getcheckboxpixmap(QStyle.State_On,
+                                                   Qt.gray, self)
+        self.selected = self.sci.markerDefine(self._checkedpix, -1)
+
+        self._uncheckedpix = qtlib.getcheckboxpixmap(QStyle.State_Off,
+                                                     Qt.gray, self)
+        self.unselected = self.sci.markerDefine(self._uncheckedpix, -1)
+
         self.vertical = self.sci.markerDefine(qsci.VerticalLine, -1)
         self.divider = self.sci.markerDefine(qsci.Background, -1)
         self.selcolor = self.sci.markerDefine(qsci.Background, -1)
@@ -659,7 +665,25 @@ class DiffBrowser(QFrame):
         self.countselected = 0
         self.updateSummary()
 
-    def displayFile(self, filename, status):
+    def _setupForceViewIndicator(self):
+        if not self._forceviewindicator:
+            self._forceviewindicator = self.sci.indicatorDefine(self.sci.PlainIndicator)
+            self.sci.setIndicatorDrawUnder(True, self._forceviewindicator)
+            self.sci.setIndicatorForegroundColor(
+                QColor('blue'), self._forceviewindicator)
+            # delay until next event-loop in order to complete mouse release
+            self.sci.SCN_INDICATORRELEASE.connect(self.forceDisplayFile,
+                                                  Qt.QueuedConnection)
+
+    def forceDisplayFile(self):
+        if self.curchunks:
+            return
+        self.sci.setText(_('Please wait while the file is opened ...'))
+        QTimer.singleShot(10,
+            lambda: self.displayFile(self._lastfile, self._status, force=True))
+
+    def displayFile(self, filename, status, force=False):
+        self._status = status
         self.clearDisplay()
         if filename == self._lastfile:
             reenable = [(c.fromline, len(c.before)) for c in self.curchunks[1:]\
@@ -669,7 +693,7 @@ class DiffBrowser(QFrame):
         self._lastfile = filename
         self.clearChunks()
 
-        fd = filedata.FileData(self._ctx, None, filename, status)
+        fd = filedata.FileData(self._ctx, None, filename, status, force=force)
 
         if fd.elabel:
             self.extralabel.setText(fd.elabel)
@@ -679,7 +703,18 @@ class DiffBrowser(QFrame):
         self.filenamelabel.setText(fd.flabel)
 
         if not fd.isValid() or not fd.diff:
-            self.sci.setText(fd.error or '')
+            if fd.error is None:
+                self.sci.clear()
+                return
+            self.sci.setText(fd.error)
+            forcedisplaymsg = filedata.forcedisplaymsg
+            linkstart = fd.error.find(forcedisplaymsg)
+            if linkstart >= 0:
+                # add the link to force to view the data anyway
+                self._setupForceViewIndicator()
+                self.sci.fillIndicatorRange(
+                    0, linkstart, 0, linkstart+len(forcedisplaymsg),
+                    self._forceviewindicator)
             return
         elif type(self._ctx.rev()) is str:
             chunks = self._ctx._files[filename]
@@ -701,10 +736,10 @@ class DiffBrowser(QFrame):
         self.sci.markerDeleteAll(-1)
         for chunk in chunks[1:]:
             chunk.lrange = (start, start+len(chunk.lines))
-            chunk.mline = start + len(chunk.lines)/2
+            chunk.mline = start
             if start:
                 self.sci.markerAdd(start-1, self.divider)
-            for i in xrange(1,len(chunk.lines)-1):
+            for i in xrange(0,len(chunk.lines)):
                 if start + i == chunk.mline:
                     self.sci.markerAdd(chunk.mline, self.unselected)
                 else:
