@@ -7,21 +7,54 @@
 
 import os
 import re
+import tempfile
 
 from mercurial import ui, util, error, scmutil, phases
 
-from tortoisehg.util import hglib, shlib, wconfig, bugtraq
+from tortoisehg.util import hglib, shlib, wconfig
 
 from tortoisehg.hgqt.i18n import _
 from tortoisehg.hgqt.messageentry import MessageEntry
 from tortoisehg.hgqt import qtlib, qscilib, status, cmdui, branchop, revpanel
-from tortoisehg.hgqt import hgrcutil, mq, lfprompt
-from tortoisehg.util.hgversion import hgversion
+from tortoisehg.hgqt import hgrcutil, mqutil, lfprompt, i18n, partialcommit
 
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 from PyQt4.Qsci import QsciAPIs
 
+if os.name == 'nt':
+    from tortoisehg.util import bugtraq
+    _hasbugtraq = True
+else:
+    _hasbugtraq = False
+
+def readrepoopts(repo):
+    opts = {}
+    opts['ciexclude'] = repo.ui.config('tortoisehg', 'ciexclude', '')
+    opts['pushafter'] = repo.ui.config('tortoisehg', 'cipushafter', '')
+    opts['autoinc'] = repo.ui.config('tortoisehg', 'autoinc', '')
+    opts['recurseinsubrepos'] = repo.ui.config('tortoisehg', 'recurseinsubrepos', None)
+    opts['bugtraqplugin'] = repo.ui.config('tortoisehg', 'issue.bugtraqplugin', None)
+    opts['bugtraqparameters'] = repo.ui.config('tortoisehg', 'issue.bugtraqparameters', None)
+    if opts['bugtraqparameters']:
+        opts['bugtraqparameters'] = os.path.expandvars(opts['bugtraqparameters'])
+    opts['bugtraqtrigger'] = repo.ui.config('tortoisehg', 'issue.bugtraqtrigger', None)
+
+    return opts
+
+def commitopts2str(opts, mode='commit'):
+    optslist = []
+    for opt, value in opts.iteritems():
+        if opt in ['user', 'date', 'pushafter', 'autoinc',
+                   'recurseinsubrepos']:
+            if mode == 'merge' and opt == 'autoinc':
+                # autoinc does not apply to merge commits
+                continue
+            if value is True:
+                optslist.append('--' + opt)
+            elif value:
+                optslist.append('--%s=%s' % (opt, value))
+    return ' '.join(optslist)
 
 # Technical Debt for CommitWidget
 #  disable commit button while no message is entered or no files are selected
@@ -32,7 +65,6 @@ from PyQt4.Qsci import QsciAPIs
 class CommitWidget(QWidget, qtlib.TaskWidget):
     'A widget that encompasses a StatusWidget and commit extras'
     commitButtonEnable = pyqtSignal(bool)
-    mqButtonEnable = pyqtSignal(bool)
     linkActivated = pyqtSignal(QString)
     showMessage = pyqtSignal(unicode)
     commitComplete = pyqtSignal()
@@ -44,7 +76,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
     endSuppressPrompt = pyqtSignal()
 
     def __init__(self, repo, pats, opts, embedded=False, parent=None, rev=None):
-        QWidget.__init__(self, parent=parent)
+        QWidget.__init__(self, parent)
 
         repo.configChanged.connect(self.configChanged)
         repo.repositoryChanged.connect(self.repositoryChanged)
@@ -56,16 +88,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.currentAction = None
         self.currentProgress = None
 
-        opts['ciexclude'] = repo.ui.config('tortoisehg', 'ciexclude', '')
-        opts['pushafter'] = repo.ui.config('tortoisehg', 'cipushafter', '')
-        opts['autoinc'] = repo.ui.config('tortoisehg', 'autoinc', '')
-        opts['recurseinsubrepos'] = repo.ui.config('tortoisehg', 'recurseinsubrepos', None)
-        opts['bugtraqplugin'] = repo.ui.config('tortoisehg', 'issue.bugtraqplugin', None)
-        opts['bugtraqparameters'] = repo.ui.config('tortoisehg', 'issue.bugtraqparameters', None)
-        if opts['bugtraqparameters']:
-            opts['bugtraqparameters'] = os.path.expandvars(opts['bugtraqparameters'])
-        opts['bugtraqtrigger'] = repo.ui.config('tortoisehg', 'issue.bugtraqtrigger', None)
-        self.opts = opts # user, date
+        self.opts = opts = readrepoopts(repo) # user, date
 
         self.stwidget = status.StatusWidget(repo, pats, opts, self)
         self.stwidget.showMessage.connect(self.showMessage)
@@ -118,7 +141,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         tbar.addAction(_('Options')).triggered.connect(self.details)
         tbar.setIconSize(QSize(16,16))
 
-        if self.opts['bugtraqplugin'] != None:
+        if _hasbugtraq and self.opts['bugtraqplugin'] != None:
             # We create the "Show Issues" button, but we delay its setup
             # because creating the bugtraq object is slow and blocks the GUI,
             # which would result in a noticeable slow down while creating the commit widget
@@ -162,7 +185,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             pnhbox = QHBoxLayout()
             self.pnlabel = QLabel()
             pnhbox.addWidget(self.pnlabel)
-            self.pnedit = mq.getPatchNameLineEdit()
+            self.pnedit = mqutil.getPatchNameLineEdit()
             self.pnedit.setMaximumWidth(250)
             pnhbox.addWidget(self.pnedit)
             pnhbox.addStretch()
@@ -196,6 +219,8 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         upperframe.setLayout(vbox)
 
         self.split = QSplitter(Qt.Vertical)
+        if os.name == 'nt':
+            self.split.setStyle(QStyleFactory.create('Plastique'))
         sp = SP(SP.Expanding, SP.Expanding)
         sp.setHorizontalStretch(1)
         sp.setVerticalStretch(0)
@@ -210,11 +235,6 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         # add our splitter where the docf used to be
         self.stwidget.split.addWidget(self.split)
         self.msgte = msgte
-        if not self.hasmqbutton:
-            QShortcut(QKeySequence('Ctrl+Return'), self,
-                      self.commit).setContext(Qt.WidgetWithChildrenShortcut)
-            QShortcut(QKeySequence('Ctrl+Enter'), self,
-                      self.commit).setContext(Qt.WidgetWithChildrenShortcut)
 
     @property
     def rev(self):
@@ -236,7 +256,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             preferredActionName = self._getPreferredActionName()
             curractionName = self.mqgroup.checkedAction()._name
             if curractionName != preferredActionName:
-                self.mqSetAction(refresh=True,
+                self.commitSetAction(refresh=True,
                     actionName=preferredActionName)
 
     def _getPreferredActionName(self):
@@ -256,29 +276,30 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                 # Set the button to QRefresh
                 return 'qref'
 
-    def mqSetupButton(self):
+    def commitSetupButton(self):
         ispatch = lambda r: 'qtip' in r.changectx('.').tags()
         notpatch = lambda r: 'qtip' not in r.changectx('.').tags()
-        canamend = lambda r: False
-        # hg >= 2.2 has amend capabilities
-        if hgversion >= '2.2':
-            def canamend(r):
-                if ispatch(r):
-                    return False
-                ctx = r.changectx('.')
-                return not ctx.children() \
-                    and ctx.phase() != phases.public \
-                    and len(ctx.parents()) < 2 \
-                    and len(r.changectx(None).parents()) < 2
+        def canamend(r):
+            if ispatch(r):
+                return False
+            ctx = r.changectx('.')
+            return not ctx.children() \
+                and ctx.phase() != phases.public \
+                and len(ctx.parents()) < 2 \
+                and len(r.changectx(None).parents()) < 2
 
-        acts = (
+        acts = [
             ('commit', _('Commit changes'), _('Commit'), notpatch),
             ('amend', _('Amend current revision'), _('Amend'), canamend),
-            ('qnew', _('Create a new patch'), _('QNew'), None),
-            ('qref', _('Refresh current patch'), _('QRefresh'), ispatch),
-        )
+        ]
+        if self.hasmqbutton:
+            acts += [
+                ('qnew', _('Create a new patch'), _('QNew'), None),
+                ('qref', _('Refresh current patch'), _('QRefresh'), ispatch),
+            ]
+        acts = tuple(acts)
 
-        class MQToolButton(QToolButton):
+        class CommitToolButton(QToolButton):
             def styleOption(self):
                 opt = QStyleOptionToolButton()
                 opt.initFrom(self)
@@ -298,13 +319,13 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                 # Set the desired width to keep the button from resizing
                 return QSize(self._width, QToolButton.sizeHint(self).height())
 
-        self.mqtb = mqtb = MQToolButton(self)
-        mqtb.setBold()
-        mqtb.setPopupMode(QToolButton.MenuButtonPopup)
-        fmk = lambda s: mqtb.fontMetrics().width(hglib.tounicode(s[2]))
-        mqtb._width = max(map(fmk, acts)) + 4*mqtb.menuButtonWidth()
+        self.committb = committb = CommitToolButton(self)
+        committb.setBold()
+        committb.setPopupMode(QToolButton.MenuButtonPopup)
+        fmk = lambda s: committb.fontMetrics().width(hglib.tounicode(s[2]))
+        committb._width = max(map(fmk, acts)) + 4*committb.menuButtonWidth()
 
-        class MQMenu(QMenu):
+        class CommitButtonMenu(QMenu):
             def __init__(self, parent, repo):
                 self.repo = repo
                 return QMenu.__init__(self, parent)
@@ -316,8 +337,8 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                         a.setEnabled(a._enablefunc(self.repo))
                 return QMenu.showEvent(self, event)
         self.mqgroup = QActionGroup(self)
-        mqmenu = MQMenu(mqtb, self.repo)
-        menurefresh = lambda: self.mqSetAction(refresh=True)
+        commitbmenu = CommitButtonMenu(committb, self.repo)
+        menurefresh = lambda: self.commitSetAction(refresh=True)
         for a in acts:
             action = QAction(a[1], self.mqgroup)
             action._name = a[0]
@@ -325,19 +346,20 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             action._enablefunc = a[3]
             action.triggered.connect(menurefresh)
             action.setCheckable(True)
-            mqmenu.addAction(action)
-        mqtb.setMenu(mqmenu)
-        mqtb.clicked.connect(self.mqPerformAction)
-        self.mqButtonEnable.connect(mqtb.setEnabled)
-        self.mqSetAction(actionName=self._getPreferredActionName())
+            commitbmenu.addAction(action)
+        committb.setMenu(commitbmenu)
+        committb.clicked.connect(self.mqPerformAction)
+        self.commitButtonEnable.connect(committb.setEnabled)
+        self.commitSetAction(actionName=self._getPreferredActionName())
         sc = QShortcut(QKeySequence('Ctrl+Return'), self, self.mqPerformAction)
         sc.setContext(Qt.WidgetWithChildrenShortcut)
         sc = QShortcut(QKeySequence('Ctrl+Enter'), self, self.mqPerformAction)
         sc.setContext(Qt.WidgetWithChildrenShortcut)
-        return mqtb
+        return committb
 
     @pyqtSlot(bool)
-    def mqSetAction(self, refresh=False, actionName=None):
+    def commitSetAction(self, refresh=False, actionName=None):
+        allowcs = False
         if actionName:
             selectedAction = \
                 [act for act in self.mqgroup.actions() \
@@ -350,17 +372,18 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             self.pnlabel.setVisible(True)
             self.pnedit.setVisible(True)
             self.pnedit.setFocus()
-            self.pnedit.setText(mq.defaultNewPatchName(self.repo))
+            self.pnedit.setText(mqutil.defaultNewPatchName(self.repo))
             self.pnedit.selectAll()
             self.stwidget.setPatchContext(None)
             refreshwctx = refresh and oldpctx is not None
         else:
-            self.pnlabel.setVisible(False)
-            self.pnedit.setVisible(False)
+            if self.hasmqbutton:
+                self.pnlabel.setVisible(False)
+                self.pnedit.setVisible(False)
             ispatch = 'qtip' in pctx.tags()
             def switchAction(action, name):
                 action.setChecked(False)
-                action = self.mqtb.menu().getActionByName(name)
+                action = self.committb.menu().getActionByName(name)
                 action.setChecked(True)
                 return action
             if curraction._name == 'qref' and not ispatch:
@@ -373,6 +396,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             elif curraction._name == 'commit':
                 refreshwctx = refresh and oldpctx is not None
                 self.stwidget.setPatchContext(None)
+                allowcs = len(self.repo.parents()) == 1
         if curraction._name in ('qref', 'amend'):
             if self.lastAction not in ('qref', 'amend'):
                 self.lastCommitMsg = self.msgte.text()
@@ -380,9 +404,12 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         else:
             if self.lastAction in ('qref', 'amend'):
                 self.setMessage(self.lastCommitMsg)
+        self.stwidget.fileview.enableChangeSelection(allowcs)
+        if not allowcs:
+            self.stwidget.partials = {}
         if refreshwctx:
             self.stwidget.refreshWctx()
-        self.mqtb.setText(curraction._text)
+        self.committb.setText(curraction._text)
         self.lastAction = curraction._name
 
     def getBranchCommandLine(self, branchName, repo):
@@ -439,30 +466,40 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             return self.commit(amend=True)
 
         # Check if we need to change branch first
-        commandlines = []
+        wholecmdlines = []  # [[cmd1, ...], [cmd2, ...], ...]
         if self.branchop:
-            commandlines, newbranch = self.getBranchCommandLine(self.branchop,
-                                                                self.repo)
-            if commandlines is None:
+            cmdlines, newbranch = self.getBranchCommandLine(self.branchop,
+                                                            self.repo)
+            if cmdlines is None:
                 return
+            wholecmdlines.extend(cmdlines)
+
         olist = ('user', 'date')
-        cmdlines = commandlines + mq.mqNewRefreshCommand(self.repo,
-                                          curraction._name == 'qnew',
-                                          self.stwidget, self.pnedit,
-                                          self.msgte.text(), self.opts, olist)
+        cmdlines = mqutil.mqNewRefreshCommand(self.repo,
+                                              curraction._name == 'qnew',
+                                              self.stwidget, self.pnedit,
+                                              self.msgte.text(), self.opts,
+                                              olist)
+        if not cmdlines:
+            return
+        wholecmdlines.extend(cmdlines)
+
         self.repo.incrementBusyCount()
         self.currentAction = curraction._name
         self.currentProgress = _('MQ Action', 'start progress')
         self.progress.emit(*cmdui.startProgress(self.currentProgress, ''))
         self.commitButtonEnable.emit(False)
-        self.mqButtonEnable.emit(False)
-        self.runner.run(*cmdlines)
+        self.runner.run(*wholecmdlines)
 
     @pyqtSlot(QString, QString)
     def fileDisplayed(self, wfile, contents):
         'Status widget is displaying a new file'
         if not (wfile and contents):
             return
+        if self.msgte.autoCompletionThreshold() < 0:
+            # do not search for tokens if auto completion is disabled
+            # pygments has several infinite loop problems we'd like to avoid
+            return
         wfile = unicode(wfile)
         self._apis = QsciAPIs(self.msgte.lexer())
         tokens = set()
@@ -496,7 +533,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.msgte.lexer().setAPIs(self._apis)
 
     def bugTrackerPostCommit(self):
-        if self.opts['bugtraqtrigger'] != 'commit':
+        if not _hasbugtraq or self.opts['bugtraqtrigger'] != 'commit':
             return
         # commit already happened, get last message in history
         message = self.lastmessage
@@ -518,7 +555,10 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.setMessage(newMessage)
 
     def details(self):
-        dlg = DetailsDialog(self.opts, self.userhist, self)
+        mode = 'commit'
+        if len(self.repo.parents()) > 1:
+            mode = 'merge'
+        dlg = DetailsDialog(self.opts, self.userhist, self, mode=mode)
         dlg.finished.connect(dlg.deleteLater)
         dlg.setWindowFlags(Qt.Sheet)
         dlg.setWindowModality(Qt.WindowModal)
@@ -526,15 +566,18 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
             self.opts.update(dlg.outopts)
             self.refresh()
 
+    @pyqtSlot()
     def workingBranchChanged(self):
         'Repository has detected a change in .hg/branch'
         self.refresh()
 
+    @pyqtSlot()
     def repositoryChanged(self):
         'Repository has detected a changelog / dirstate change'
         self.refresh()
         self.stwidget.refreshWctx() # Trigger reload of working context
 
+    @pyqtSlot()
     def configChanged(self):
         'Repository is reporting its config files have changed'
         self.refresh()
@@ -553,7 +596,8 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
 
     def refresh(self):
         ispatch = self.repo.changectx('.').thgmqappliedpatch()
-        self.commitButtonEnable.emit(not ispatch)
+        if not self.hasmqbutton:
+            self.commitButtonEnable.emit(not ispatch)
         self.msgte.refresh(self.repo)
 
         # Update branch operation button
@@ -567,18 +611,10 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.branchbutton.setText(title)
 
         # Update options label, showing only whitelisted options.
-        opts = []
-        for opt, value in self.opts.iteritems():
-            if opt in ['user', 'date', 'pushafter', 'autoinc',
-                       'recurseinsubrepos']:
-                if value is True:
-                    opts.append('--' + opt)
-                elif value:
-                    opts.append('--%s=%s' % (opt, value))
-
+        opts = commitopts2str(self.opts)
         self.optionslabelfmt = _('<b>Selected Options:</b> %s')
         self.optionslabel.setText(self.optionslabelfmt
-                                  % hglib.tounicode(' '.join(opts)))
+                                  % hglib.tounicode(opts))
         self.optionslabel.setVisible(bool(opts))
 
         # Update parent csinfo widget
@@ -601,7 +637,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                 self.pnlabel.setText(start + text[firstidx+secondidx:])
             else:
                 self.pnlabel.setText(patchname)
-            self.mqSetAction()
+            self.commitSetAction()
 
     def branchOp(self):
         d = branchop.BranchOpDialog(self.repo, self.branchop, self)
@@ -609,6 +645,14 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         d.setWindowModality(Qt.WindowModal)
         if d.exec_() == QDialog.Accepted:
             self.branchop = d.branchop
+            if self.branchop is False:
+                if not self.getMessage(True).strip():
+                    engmsg = self.repo.ui.configbool(
+                        'tortoisehg', 'engmsg', False)
+                    msgset = i18n.keepgettext()._('Close %s branch')
+                    text = engmsg and msgset['id'] or msgset['str']
+                    self.setMessage(unicode(text) %
+                                    hglib.tounicode(self.repo[None].branch()))
             self.refresh()
 
     def canUndo(self):
@@ -635,8 +679,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.currentProgress = _('Rollback', 'start progress')
         self.progress.emit(*cmdui.startProgress(self.currentProgress, ''))
         self.commitButtonEnable.emit(False)
-        self.mqButtonEnable.emit(False)
-        self.runner.run(['rollback'])
+        self.runner.run(['rollback', '--repository', self.repo.root])
         self.stopAction.setEnabled(True)
 
     def updateRecentMessages(self):
@@ -706,18 +749,18 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
 
     def saveSettings(self, s, prefix):
         'Save history, etc, in QSettings instance'
-        repoid = str(self.repo[0])
-        lpref = prefix + '/commit/'
-        gpref = 'commit/'
-        s.setValue(lpref+'split', self.split.saveState())
-        self.msgte.saveSettings(s, lpref+'msgte')
-        self.stwidget.saveSettings(s, lpref+'status')
-        s.setValue(gpref+'history-'+repoid, self.msghistory)
-        s.setValue(gpref+'userhist', self.userhist)
-        msg = self.getMessage(True)
         try:
+            repoid = str(self.repo[0])
+            lpref = prefix + '/commit/'
+            gpref = 'commit/'
+            s.setValue(lpref+'split', self.split.saveState())
+            self.msgte.saveSettings(s, lpref+'msgte')
+            self.stwidget.saveSettings(s, lpref+'status')
+            s.setValue(gpref+'history-'+repoid, self.msghistory)
+            s.setValue(gpref+'userhist', self.userhist)
+            msg = self.getMessage(True)
             self.repo.opener('cur-message.txt', 'w').write(msg)
-        except EnvironmentError:
+        except (EnvironmentError, IOError):
             pass
 
     def addMessageToHistory(self, umsg):
@@ -748,6 +791,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                      (_('&Replace'), _('Cancel')), 0, 1, []).run()
             if res == 0:
                 msg = self.getMessage(True)
+                msg = str(msg)  # drop round-trip utf8 data
                 self.msgte.setText(hglib.tounicode(msg))
             self.msgte.setFocus()
             return
@@ -788,15 +832,25 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                                                                 self.repo)
             if commandlines is None:
                 return
-        self.files = self.stwidget.getChecked('MAR?!S')
-        if not (self.files or brcmd or newbranch):
+        partials = []
+        if len(repo.parents()) > 1:
+            merge = True
+            self.files = []
+        else:
+            merge = False
+            files = self.stwidget.getChecked('MAR?!S')
+            # make list of files with partial change selections
+            for fname, c in self.stwidget.partials.iteritems():
+                if c.excludecount > 0 and c.excludecount < len(c.hunks):
+                    partials.append(fname)
+            self.files = set(files + partials)
+        canemptycommit = bool(brcmd or newbranch or amend)
+        if not (self.files or canemptycommit or merge):
             qtlib.WarningMsgBox(_('No files checked'),
                                 _('No modified files checkmarked for commit'),
                                 parent=self)
             self.stwidget.tv.setFocus()
             return
-        if len(repo.parents()) > 1:
-            self.files = []
 
         user = qtlib.getCurrentUsername(self, self.repo, self.opts)
         if not user:
@@ -867,12 +921,35 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
                    '--user', user, '--message='+msg]
         cmdline += dcmd + brcmd
 
+        if partials:
+            partialcommit.uisetup(repo.ui)
+
+            # write patch for partial change selections to temp file
+            fd, tmpname = tempfile.mkstemp(prefix='thg-patch-')
+            fp = os.fdopen(fd, 'wb')
+            for fname in partials:
+                changes = self.stwidget.partials[fname]
+                changes.write(fp)
+                for chunk in changes.hunks:
+                    if not chunk.excluded:
+                        chunk.write(fp)
+            fp.close()
+
+            cmdline.append('--partials')
+            cmdline.append(tmpname)
+            assert not amend
+
         if self.opts.get('recurseinsubrepos'):
             cmdline.append('--subrepos')
 
         if amend:
             cmdline.append('--amend')
 
+        if not self.files and canemptycommit and not merge:
+            # make sure to commit empty changeset by excluding all files
+            cmdline.extend(['--exclude', repo.root])
+            assert not self.stwidget.partials
+
         cmdline.append('--')
         cmdline.extend([repo.wjoin(f) for f in self.files])
         if len(repo.parents()) == 1:
@@ -884,6 +961,8 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
 
         if self.opts.get('pushafter'):
             cmd = ['push', '--repository', repo.root, self.opts['pushafter']]
+            if newbranch:
+                cmd.append('--new-branch')
             commandlines.append(cmd)
 
         repo.incrementBusyCount()
@@ -894,7 +973,6 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.currentProgress = _('Commit', 'start progress')
         self.progress.emit(*cmdui.startProgress(self.currentProgress, ''))
         self.commitButtonEnable.emit(False)
-        self.mqButtonEnable.emit(False)
         self.runner.run(*commandlines)
         self.stopAction.setEnabled(True)
 
@@ -905,9 +983,9 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
         self.progress.emit(*cmdui.stopProgress(self.currentProgress))
         self.stopAction.setEnabled(False)
         self.commitButtonEnable.emit(True)
-        self.mqButtonEnable.emit(True)
         self.repo.decrementBusyCount()
         if ret == 0:
+            self.stwidget.partials = {}
             if self.currentAction == 'rollback':
                 shlib.shell_notify([self.repo.root])
                 return
@@ -927,7 +1005,7 @@ class CommitWidget(QWidget, qtlib.TaskWidget):
 
 class DetailsDialog(QDialog):
     'Utility dialog for configuring uncommon settings'
-    def __init__(self, opts, userhistory, parent):
+    def __init__(self, opts, userhistory, parent, mode='commit'):
         QDialog.__init__(self, parent)
         self.setWindowTitle(_('%s - commit options') % parent.repo.displayname)
         self.repo = parent.repo
@@ -1046,8 +1124,10 @@ class DetailsDialog(QDialog):
         hbox.addWidget(self.autoinccb)
         hbox.addWidget(self.autoincle)
         hbox.addWidget(autoincsave)
-        layout.addLayout(hbox)
-        
+        if mode != 'merge':
+            #self.autoinccb.setVisible(False)
+            layout.addLayout(hbox)
+
         hbox = QHBoxLayout()
         recursesave = QPushButton(_('Save in Repo'))
         recursesave.clicked.connect(self.saveRecurseInSubrepos)
@@ -1055,14 +1135,14 @@ class DetailsDialog(QDialog):
         SP = QSizePolicy
         self.recursecb.setSizePolicy(SP(SP.Expanding, SP.Minimum))
         #self.recursecb.toggled.connect(recursesave.setEnabled)
-        
+
         if opts.get('recurseinsubrepos'):
             self.recursecb.setChecked(True)
-            
+
         hbox.addWidget(self.recursecb)
         hbox.addWidget(recursesave)
         layout.addLayout(hbox)
-        
+
         BB = QDialogButtonBox
         bb = QDialogButtonBox(BB.Ok|BB.Cancel)
         bb.accepted.connect(self.accept)
@@ -1219,7 +1299,7 @@ class DetailsDialog(QDialog):
             outopts['recurseinsubrepos'] = 'true'
         else:
             outopts['recurseinsubrepos'] = ''
-        
+
         self.outopts = outopts
         QDialog.accept(self)
 
@@ -1257,18 +1337,8 @@ class CommitDialog(QDialog):
         bb.button(BB.Discard).clicked.connect(commit.rollback)
         bb.button(BB.Close).setDefault(False)
         bb.button(BB.Discard).setDefault(False)
-        if commit.hasmqbutton:
-            self.commitButton = commit.mqSetupButton()
-            bb.addButton(self.commitButton, BB.AcceptRole)
-        else:
-            self.commitButton = commitbtn = bb.addButton(BB.Ok)
-            commitbtn.setDefault(True)
-            commitbtn.setText(_('Commit', 'action button'))
-            f = commitbtn.font()
-            f.setWeight(QFont.Bold)
-            commitbtn.setFont(f)
-            commit.commitButtonEnable.connect(commitbtn.setEnabled)
-            bb.accepted.connect(self.accept)
+        self.commitButton = commit.commitSetupButton()
+        bb.addButton(self.commitButton, BB.AcceptRole)
 
         self.bb = bb
 
@@ -1286,7 +1356,7 @@ class CommitDialog(QDialog):
         self.commit.reload()
         self.updateUndo()
         self.commit.msgte.setFocus()
-        QShortcut(QKeySequence.Refresh, self, self.refresh)
+        qtlib.newshortcutsforstdkey(QKeySequence.Refresh, self, self.refresh)
 
     def done(self, ret):
         self.commit.repo.configChanged.disconnect(self.commit.configChanged)
@@ -1297,9 +1367,9 @@ class CommitDialog(QDialog):
 
     def linkActivated(self, link):
         link = hglib.fromunicode(link)
-        if link.startswith('subrepo:'):
+        if link.startswith('repo:'):
             from tortoisehg.hgqt.run import qtrun
-            qtrun(run, ui.ui(), root=link[8:])
+            qtrun(run, ui.ui(), root=link[len('repo:'):])
         if link.startswith('shelve:'):
             repo = self.commit.repo
             from tortoisehg.hgqt import shelve
@@ -1342,7 +1412,7 @@ class CommitDialog(QDialog):
             s.setValue('commit/geom', self.saveGeometry())
             self.commit.saveSettings(s, 'committool')
         return exit
-    
+
     def accept(self):
         self.commit.commit()
 
@@ -1350,10 +1420,6 @@ class CommitDialog(QDialog):
         if self.promptExit():
             QDialog.reject(self)
 
-    def closeEvent(self, event):
-        if not self.promptExit():
-            event.ignore()
-
 def run(ui, *pats, **opts):
     from tortoisehg.util import paths
     from tortoisehg.hgqt import thgrepo
