@@ -24,6 +24,7 @@ from tortoisehg.util import hglib
 from tortoisehg.hgqt.i18n import _
 from tortoisehg.hgqt import qtlib, visdiff, filerevmodel, blockmatcher, lexers
 from tortoisehg.hgqt import fileview, repoview, revpanel, revert
+from tortoisehg.hgqt.qscilib import Scintilla
 
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
@@ -33,6 +34,8 @@ sides = ('left', 'right')
 otherside = {'left': 'right', 'right': 'left'}
 
 class _AbstractFileDialog(QMainWindow):
+    finished = pyqtSignal(int)
+
     def __init__(self, repo, filename, repoviewer=None):
         QMainWindow.__init__(self)
         self.repo = repo
@@ -51,10 +54,18 @@ class _AbstractFileDialog(QMainWindow):
         self.setupViews()
         self.setupModels()
 
+    def closeEvent(self, event):
+        super(_AbstractFileDialog, self).closeEvent(event)
+        self.finished.emit(0)  # mimic QDialog exit
+
     def setRepoViewer(self, repoviewer=None):
         self.repoviewer = repoviewer
         if repoviewer:
-            repoviewer.finished.connect(lambda x: self.setRepoViewer(None))
+            repoviewer.finished.connect(self._clearRepoViewer)
+
+    @pyqtSlot()
+    def _clearRepoViewer(self):
+        self.setRepoViewer(None)
 
     def reload(self):
         'Reload toolbar action handler'
@@ -83,6 +94,7 @@ class FileLogDialog(_AbstractFileDialog):
         super(FileLogDialog, self).__init__(repo, filename, repoviewer)
         self._readSettings()
         self.menu = None
+        self.dualmenu = None
 
     def closeEvent(self, event):
         self._writeSettings()
@@ -114,8 +126,10 @@ class FileLogDialog(_AbstractFileDialog):
         self.editToolbar = QToolBar(self)
         self.editToolbar.setContextMenuPolicy(Qt.PreventContextMenu)
         self.addToolBar(Qt.ToolBarArea(Qt.TopToolBarArea), self.editToolbar)
-        self.actionClose = QAction(self, shortcut=QKeySequence.Close)
-        self.actionReload = QAction(self, shortcut=QKeySequence.Refresh)
+        self.actionClose = QAction(self)
+        self.actionClose.setShortcuts(QKeySequence.Close)
+        self.actionReload = QAction(self)
+        self.actionReload.setShortcuts(QKeySequence.Refresh)
         self.editToolbar.addAction(self.actionReload)
         self.addAction(self.actionClose)
 
@@ -123,6 +137,9 @@ class FileLogDialog(_AbstractFileDialog):
         self.setCentralWidget(self.splitter)
         cs = ('fileLogDialog', _('File History Log Columns'))
         self.repoview = repoview.HgRepoView(self.repo, cs[0], cs, self.splitter)
+        self.repoview.revisionSelectionChanged.connect(
+            self._checkValidSelection)
+
         self.contentframe = QFrame(self.splitter)
 
         vbox = QVBoxLayout()
@@ -189,35 +206,49 @@ class FileLogDialog(_AbstractFileDialog):
     @pyqtSlot(QPoint, object)
     def viewMenuRequest(self, point, selection):
         'User requested a context menu in repo view widget'
-        if not selection:
+        if not selection or len(selection) > 2:
             return
-        if self.menu is None:
+        if len(selection) == 2:
+            if self.dualmenu is None:
+                self.dualmenu = menu = QMenu(self)
+                a = menu.addAction(_('&Diff Selected Changesets'))
+                a.triggered.connect(self.onVisualDiffRevs)
+                a = menu.addAction(_('Diff Selected &File Revisions'))
+                a.setIcon(qtlib.getmenuicon('visualdiff'))
+                a.triggered.connect(self.onVisualDiffFileRevs)
+            else:
+                menu = self.dualmenu
+        elif self.menu is None:
             self.menu = menu = QMenu(self)
-            a = menu.addAction(_('Visual diff...'))
+            a = menu.addAction(_('&Diff Changeset to Parent'))
             a.setIcon(qtlib.getmenuicon('visualdiff'))
             a.triggered.connect(self.onVisualDiff)
-            a = menu.addAction(_('Diff to local...'))
+            a = menu.addAction(_('Diff Changeset to &Local'))
             a.setIcon(qtlib.getmenuicon('ldiff'))
             a.triggered.connect(self.onVisualDiffToLocal)
             menu.addSeparator()
-            a = menu.addAction(_('Visual diff file...'))
+            a = menu.addAction(_('Diff &File to Parent'))
             a.setIcon(qtlib.getmenuicon('visualdiff'))
             a.triggered.connect(self.onVisualDiffFile)
-            a = menu.addAction(_('Diff file to local...'))
+            a = menu.addAction(_('Diff File to Lo&cal'))
             a.setIcon(qtlib.getmenuicon('ldiff'))
             a.triggered.connect(self.onVisualDiffFileToLocal)
             menu.addSeparator()
-            a = menu.addAction(_('View at revision...'))
+            a = menu.addAction(_('&View at Revision'))
             a.setIcon(qtlib.getmenuicon('view-at-revision'))
             a.triggered.connect(self.onViewFileAtRevision)
-            a = menu.addAction(_('Edit local'))
+            a = menu.addAction(_('&Save at Revision...'))
+            a.triggered.connect(self.onSaveFileAtRevision)
+            a = menu.addAction(_('&Edit Local'))
             a.setIcon(qtlib.getmenuicon('edit-file'))
             a.triggered.connect(self.onEditLocal)
-            a = menu.addAction(_('Revert to revision...'))
+            a = menu.addAction(_('&Revert to Revision...'))
             a.setIcon(qtlib.getmenuicon('hg-revert'))
             a.triggered.connect(self.onRevertFileToRevision)
+        else:
+            menu = self.menu
         self.selection = selection
-        self.menu.exec_(point)
+        menu.exec_(point)
 
     def onVisualDiff(self):
         opts = dict(change=self.selection[0])
@@ -233,6 +264,17 @@ class FileLogDialog(_AbstractFileDialog):
             dlg.exec_()
             dlg.deleteLater()
 
+    def onVisualDiffRevs(self):
+        revs = self.selection
+        if len(revs) != 2:
+            self.textView.showMessage.emit(_('You must select two revisions to diff'))
+            return
+        opts = dict(rev=revs)
+        dlg = visdiff.visualdiff(self.repo.ui, self.repo, [], opts)
+        if dlg:
+            dlg.exec_()
+            dlg.deleteLater()
+
     def onVisualDiffFile(self):
         rev = self.selection[0]
         paths = [self.filerevmodel.graph.filename(rev)]
@@ -251,6 +293,18 @@ class FileLogDialog(_AbstractFileDialog):
             dlg.exec_()
             dlg.deleteLater()
 
+    def onVisualDiffFileRevs(self):
+        revs = self.selection
+        if len(revs) != 2:
+            self.textView.showMessage.emit(_('You must select two revisions to diff'))
+            return
+        paths = [self.filerevmodel.graph.filename(revs[0])]
+        opts = dict(rev=['rev(%d)' % rev for rev in revs])
+        dlg = visdiff.visualdiff(self.repo.ui, self.repo, paths, opts)
+        if dlg:
+            dlg.exec_()
+            dlg.deleteLater()
+
     def onEditLocal(self):
         filenames = [self.filename]
         if not filenames:
@@ -282,6 +336,14 @@ class FileLogDialog(_AbstractFileDialog):
                      for filename in filenames]
             qtlib.editfiles(self.repo, files, parent=self)
 
+    def onSaveFileAtRevision(self, rev):
+        rev = self.selection[0]
+        files = [self.filerevmodel.graph.filename(rev)]
+        if not files or rev is None:
+            return
+        else:
+            qtlib.savefiles(self.repo, files, rev, parent=self)
+
     @pyqtSlot(QString)
     def onLinkActivated(self, link):
         link = unicode(link)
@@ -301,6 +363,17 @@ class FileLogDialog(_AbstractFileDialog):
         self.revpanel.set_revision(rev)
         self.revpanel.update(repo = self.repo)
 
+    # It does not make sense to select more than two revisions at a time.
+    # Rather than enforcing a max selection size we simply let the user
+    # know when it has selected too many revisions by using the status bar
+    def _checkValidSelection(self, selected, deselected):
+        selection = self.repoview.selectedRevisions()
+        if len(selection) > 2:
+            msg = _('Too many rows selected for menu')
+        else:
+            msg = ''
+        self.textView.showMessage.emit(msg)
+
     def goto(self, rev):
         index = self.filerevmodel.indexFromRev(rev)
         if index is not None:
@@ -347,8 +420,10 @@ class FileDiffDialog(_AbstractFileDialog):
         self.editToolbar = QToolBar(self)
         self.editToolbar.setContextMenuPolicy(Qt.PreventContextMenu)
         self.addToolBar(Qt.ToolBarArea(Qt.TopToolBarArea), self.editToolbar)
-        self.actionClose = QAction(self, shortcut=QKeySequence.Close)
-        self.actionReload = QAction(self, shortcut=QKeySequence.Refresh)
+        self.actionClose = QAction(self)
+        self.actionClose.setShortcuts(QKeySequence.Close)
+        self.actionReload = QAction(self)
+        self.actionReload.setShortcuts(QKeySequence.Refresh)
         self.editToolbar.addAction(self.actionReload)
         self.addAction(self.actionClose)
 
@@ -368,10 +443,19 @@ class FileDiffDialog(_AbstractFileDialog):
                                                              cs, self)
         self.horizontalLayout.addWidget(self.tableView_revisions_left)
         self.horizontalLayout.addWidget(self.tableView_revisions_right)
+        self.tableView_revisions_right.setSelectionMode(QAbstractItemView.SingleSelection)
+        self.tableView_revisions_left.setSelectionMode(QAbstractItemView.SingleSelection)
         self.frame = QFrame()
         self.splitter.addWidget(layouttowidget(self.horizontalLayout))
         self.splitter.addWidget(self.frame)
 
+    #@pyqtSlot(QPoint)
+    def fileViewMenuRequest(self, point):
+        sci = self.sender()
+        menu = sci.createStandardContextMenu()
+        point = sci.viewport().mapToGlobal(point)
+        menu.exec_(point)
+
     def setupViews(self):
         self.tableViews = {'left': self.tableView_revisions_left,
                            'right': self.tableView_revisions_right}
@@ -391,11 +475,14 @@ class FileDiffDialog(_AbstractFileDialog):
             lexer = None
 
         for side, idx  in (('left', 0), ('right', 3)):
-            sci = QsciScintilla(self.frame)
+            sci = Scintilla(self.frame)
             sci.verticalScrollBar().setFocusPolicy(Qt.StrongFocus)
             sci.setFocusProxy(sci.verticalScrollBar())
             sci.verticalScrollBar().installEventFilter(self)
-            sci.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
+
+            sci.setContextMenuPolicy(Qt.CustomContextMenu)
+            sci.customContextMenuRequested.connect(self.fileViewMenuRequest)
+
             sci.setFrameShape(QFrame.NoFrame)
             sci.setMarginLineNumbers(1, True)
             sci.SendScintilla(sci.SCI_SETSELEOLFILLED, True)
@@ -439,8 +526,13 @@ class FileDiffDialog(_AbstractFileDialog):
             table.revisionSelected.connect(self.onRevisionSelected)
             table.revisionActivated.connect(self.onRevisionActivated)
 
-            self.viewers[side].verticalScrollBar().valueChanged.connect(
-                    lambda value, side=side: self.vbar_changed(value, side))
+        l, r = (self.viewers[k].verticalScrollBar() for k in sides)
+        l.valueChanged.connect(self.sbar_changed_left)
+        r.valueChanged.connect(self.sbar_changed_right)
+
+        l, r = (self.viewers[k].horizontalScrollBar() for k in sides)
+        l.valueChanged.connect(r.setValue)
+        r.valueChanged.connect(l.setValue)
 
         self.setTabOrder(table, self.viewers['left'])
         self.setTabOrder(self.viewers['left'], self.viewers['right'])
@@ -631,9 +723,17 @@ class FileDiffDialog(_AbstractFileDialog):
             self.update_page_steps(keeppos)
             self.timer.start()
 
-    def vbar_changed(self, value, side):
+    @pyqtSlot(int)
+    def sbar_changed_left(self, value):
+        self.sbar_changed(value, 'left')
+
+    @pyqtSlot(int)
+    def sbar_changed_right(self, value):
+        self.sbar_changed(value, 'right')
+
+    def sbar_changed(self, value, side):
         """
-        Callback called when the vertical scrollbar of a file viewer
+        Callback called when a scrollbar of a file viewer
         is changed, so we can update the position of the other file
         viewer.
         """
@@ -669,27 +769,29 @@ class FileDiffDialog(_AbstractFileDialog):
             return
         if self.menu is None:
             self.menu = menu = QMenu(self)
-            a = menu.addAction(_('Visual diff...'))
+            a = menu.addAction(_('&Diff to Parent'))
             a.setIcon(qtlib.getmenuicon('visualdiff'))
             a.triggered.connect(self.onVisualDiff)
-            a = menu.addAction(_('Diff to local...'))
+            a = menu.addAction(_('Diff to &Local'))
             a.setIcon(qtlib.getmenuicon('ldiff'))
             a.triggered.connect(self.onVisualDiffToLocal)
             menu.addSeparator()
-            a = menu.addAction(_('Visual diff file...'))
+            a = menu.addAction(_('Diff &File to Parent'))
             a.setIcon(qtlib.getmenuicon('visualdiff'))
             a.triggered.connect(self.onVisualDiffFile)
-            a = menu.addAction(_('Diff file to local...'))
+            a = menu.addAction(_('Diff File to Lo&cal'))
             a.setIcon(qtlib.getmenuicon('ldiff'))
             a.triggered.connect(self.onVisualDiffFileToLocal)
             menu.addSeparator()
-            a = menu.addAction(_('View at revision...'))
+            a = menu.addAction(_('&View at Revision'))
             a.setIcon(qtlib.getmenuicon('view-at-revision'))
             a.triggered.connect(self.onViewFileAtRevision)
-            a = menu.addAction(_('Edit local'))
+            a = menu.addAction(_('&Save at Revision...'))
+            a.triggered.connect(self.onSaveFileAtRevision)
+            a = menu.addAction(_('&Edit Local'))
             a.setIcon(qtlib.getmenuicon('edit-file'))
             a.triggered.connect(self.onEditLocal)
-            a = menu.addAction(_('Revert to revision...'))
+            a = menu.addAction(_('&Revert to Revision...'))
             a.setIcon(qtlib.getmenuicon('hg-revert'))
             a.triggered.connect(self.onRevertFileToRevision)
         self.selection = selection
@@ -757,3 +859,11 @@ class FileDiffDialog(_AbstractFileDialog):
             files = [os.path.join(base, filename)
                      for filename in filenames]
             qtlib.editfiles(self.repo, files, parent=self)
+
+    def onSaveFileAtRevision(self, rev):
+        rev = self.selection[0]
+        files = [self.filerevmodel.graph.filename(rev)]
+        if not files or rev is None:
+            return
+        else:
+            qtlib.savefiles(self.repo, files, rev, parent=self)
