@@ -127,8 +127,10 @@ class Scintilla(QsciScintilla):
         super(Scintilla, self).__init__(parent)
         self.autoUseTabs = True
         self.setUtf8(True)
+        self.setWrapVisualFlags(QsciScintilla.WrapFlagByBorder)
         self.textChanged.connect(self._resetfindcond)
         self._resetfindcond()
+        unbindConflictedKeys(self)
 
     def read(self, f):
         result = super(Scintilla, self).read(f)
@@ -177,48 +179,53 @@ class Scintilla(QsciScintilla):
         else:
             self._stdMenu.clear()
         if not self.isReadOnly():
-            a = self._stdMenu.addAction(_('Undo'), self.undo, QKeySequence.Undo)
+            a = self._stdMenu.addAction(_('&Undo'), self.undo)
+            a.setShortcuts(QKeySequence.Undo)
             a.setEnabled(self.isUndoAvailable())
-            a = self._stdMenu.addAction(_('Redo'), self.redo, QKeySequence.Redo)
+            a = self._stdMenu.addAction(_('&Redo'), self.redo)
+            a.setShortcuts(QKeySequence.Redo)
             a.setEnabled(self.isRedoAvailable())
             self._stdMenu.addSeparator()
-            a = self._stdMenu.addAction(_('Cut'), self.cut, QKeySequence.Cut)
+            a = self._stdMenu.addAction(_('Cu&t'), self.cut)
+            a.setShortcuts(QKeySequence.Cut)
             a.setEnabled(self.hasSelectedText())
-        a = self._stdMenu.addAction(_('Copy'), self.copy, QKeySequence.Copy)
+        a = self._stdMenu.addAction(_('&Copy'), self.copy)
+        a.setShortcuts(QKeySequence.Copy)
         a.setEnabled(self.hasSelectedText())
         if not self.isReadOnly():
-            self._stdMenu.addAction(_('Paste'), self.paste, QKeySequence.Paste)
-            a = self._stdMenu.addAction(_('Delete'), self.removeSelectedText,
-                               QKeySequence.Delete)
+            a = self._stdMenu.addAction(_('&Paste'), self.paste)
+            a.setShortcuts(QKeySequence.Paste)
+            a = self._stdMenu.addAction(_('&Delete'), self.removeSelectedText)
+            a.setShortcuts(QKeySequence.Delete)
             a.setEnabled(self.hasSelectedText())
         self._stdMenu.addSeparator()
-        self._stdMenu.addAction(_('Select All'),
-                                self.selectAll, QKeySequence.SelectAll)
+        a = self._stdMenu.addAction(_('Select &All'), self.selectAll)
+        a.setShortcuts(QKeySequence.SelectAll)
         self._stdMenu.addSeparator()
         qsci = QsciScintilla
-        wrapmenu = QMenu(_('Wrap'), self)
-        for name, mode in ((_('None', 'wrap mode'), qsci.WrapNone),
-                           (_('Word'), qsci.WrapWord),
-                           (_('Character'), qsci.WrapCharacter)):
+        wrapmenu = QMenu(_('&Wrap'), self)
+        for name, mode in ((_('&None', 'wrap mode'), qsci.WrapNone),
+                           (_('&Word'), qsci.WrapWord),
+                           (_('&Character'), qsci.WrapCharacter)):
             def mkaction(n, m):
                 a = wrapmenu.addAction(n)
                 a.setCheckable(True)
                 a.setChecked(self.wrapMode() == m)
                 a.triggered.connect(lambda: self.setWrapMode(m))
             mkaction(name, mode)
-        wsmenu = QMenu(_('Whitespace'), self)
-        for name, mode in ((_('Visible'), qsci.WsVisible),
-                           (_('Invisible'), qsci.WsInvisible),
-                           (_('AfterIndent'), qsci.WsVisibleAfterIndent)):
+        wsmenu = QMenu(_('White&space'), self)
+        for name, mode in ((_('&Visible'), qsci.WsVisible),
+                           (_('&Invisible'), qsci.WsInvisible),
+                           (_('&AfterIndent'), qsci.WsVisibleAfterIndent)):
             def mkaction(n, m):
                 a = wsmenu.addAction(n)
                 a.setCheckable(True)
                 a.setChecked(self.whitespaceVisibility() == m)
                 a.triggered.connect(lambda: self.setWhitespaceVisibility(m))
             mkaction(name, mode)
-        vsmenu = QMenu(_('EOL Visibility'), self)
-        for name, mode in ((_('Visible'), True),
-                           (_('Invisible'), False)):
+        vsmenu = QMenu(_('EOL &Visibility'), self)
+        for name, mode in ((_('&Visible'), True),
+                           (_('&Invisible'), False)):
             def mkaction(n, m):
                 a = vsmenu.addAction(n)
                 a.setCheckable(True)
@@ -229,30 +236,30 @@ class Scintilla(QsciScintilla):
         tabindentsmenu = None
         acmenu = None
         if not self.isReadOnly():
-            eolmodemenu = QMenu(_('EOL Mode'), self)
-            for name, mode in ((_('Windows'), qsci.EolWindows),
-                               (_('Unix'), qsci.EolUnix),
-                               (_('Mac'), qsci.EolMac)):
+            eolmodemenu = QMenu(_('EOL &Mode'), self)
+            for name, mode in ((_('&Windows'), qsci.EolWindows),
+                               (_('&Unix'), qsci.EolUnix),
+                               (_('&Mac'), qsci.EolMac)):
                 def mkaction(n, m):
                     a = eolmodemenu.addAction(n)
                     a.setCheckable(True)
                     a.setChecked(self.eolMode() == m)
                     a.triggered.connect(lambda: self.setEolMode(m))
                 mkaction(name, mode)
-            tabindentsmenu = QMenu(_('TAB inserts'), self)
-            for name, mode in ((_('Auto'), -1),
-                               (_('TAB'), True),
-                               (_('Spaces'), False)):
+            tabindentsmenu = QMenu(_('&TAB Inserts'), self)
+            for name, mode in ((_('&Auto'), -1),
+                               (_('&TAB'), True),
+                               (_('&Spaces'), False)):
                 def mkaction(n, m):
                     a = tabindentsmenu.addAction(n)
                     a.setCheckable(True)
                     a.setChecked(self.indentationsUseTabs() == m or \
-                        (self.autoUseTabs and n == 'Auto'))
+                        (self.autoUseTabs and m == -1))
                     a.triggered.connect(lambda: self.setIndentationsUseTabs(m))
                 mkaction(name, mode)
-            acmenu = QMenu(_('Auto-Complete'), self)
-            for name, value in ((_('Enable'), 2),
-                                (_('Disable'), -1)):
+            acmenu = QMenu(_('&Auto-Complete'), self)
+            for name, value in ((_('&Enable'), 2),
+                                (_('&Disable'), -1)):
                 def mkaction(n, v):
                     a = acmenu.addAction(n)
                     a.setCheckable(True)
@@ -260,7 +267,7 @@ class Scintilla(QsciScintilla):
                     a.triggered.connect(lambda: self.setAutoCompletionThreshold(v))
                 mkaction(name, value)
 
-        editoptsmenu = QMenu(_('Editor Options'), self)
+        editoptsmenu = QMenu(_('&Editor Options'), self)
         editoptsmenu.addMenu(wrapmenu)
         editoptsmenu.addSeparator()
         editoptsmenu.addMenu(wsmenu)
@@ -380,6 +387,56 @@ class Scintilla(QsciScintilla):
             tabs = findTabIndentsInLines(hglib.fromunicode(self.text()))
         super(Scintilla, self).setIndentationsUseTabs(tabs)
 
+    def lineNearPoint(self, point):
+        """Return the closest line to the pixel position; similar to lineAt(),
+        but returns valid line number even if no character fount at point"""
+        # lineAt() uses the strict request, SCI_POSITIONFROMPOINTCLOSE
+        chpos = self.SendScintilla(self.SCI_POSITIONFROMPOINT,
+                                   point.x(), point.y())
+        return self.SendScintilla(self.SCI_LINEFROMPOSITION, chpos)
+
+    # compability mode with QScintilla from Ubuntu 10.04
+    if not hasattr(QsciScintilla, 'HiddenIndicator'):
+        HiddenIndicator = QsciScintilla.INDIC_HIDDEN
+    if not hasattr(QsciScintilla, 'PlainIndicator'):
+        PlainIndicator = QsciScintilla.INDIC_PLAIN
+    if not hasattr(QsciScintilla, 'StrikeIndicator'):
+        StrikeIndicator = QsciScintilla.INDIC_STRIKE
+
+    if not hasattr(QsciScintilla, 'indicatorDefine'):
+        def indicatorDefine(self, style, indicatorNumber=-1):
+            # compatibility layer allows only one indicator to be defined
+            if indicatorNumber == -1:
+                indicatorNumber = 1
+            self.SendScintilla(self.SCI_INDICSETSTYLE, indicatorNumber, style)
+            return indicatorNumber
+
+    if not hasattr(QsciScintilla, 'setIndicatorDrawUnder'):
+        def setIndicatorDrawUnder(self, under, indicatorNumber):
+            self.SendScintilla(self.SCI_INDICSETUNDER, indicatorNumber, under)
+
+    if not hasattr(QsciScintilla, 'setIndicatorForegroundColor'):
+        def setIndicatorForegroundColor(self, color, indicatorNumber):
+            self.SendScintilla(self.SCI_INDICSETFORE, indicatorNumber, color);
+            self.SendScintilla(self.SCI_INDICSETALPHA, indicatorNumber, color.alpha());
+
+    if not hasattr(QsciScintilla, 'clearIndicatorRange'):
+        def clearIndicatorRange(self, lineFrom, indexFrom, lineTo, indexTo, indicatorNumber):
+            start = self.positionFromLineIndex(lineFrom, indexFrom)
+            finish = self.positionFromLineIndex(lineTo, indexTo)
+
+            self.SendScintilla(self.SCI_SETINDICATORCURRENT, indicatorNumber);
+            self.SendScintilla(self.SCI_INDICATORCLEARRANGE, start, finish - start);
+
+    if not hasattr(QsciScintilla, 'fillIndicatorRange'):
+        def fillIndicatorRange(self, lineFrom, indexFrom, lineTo, indexTo, indicatorNumber):
+            start = self.positionFromLineIndex(lineFrom, indexFrom)
+            finish = self.positionFromLineIndex(lineTo, indexTo)
+
+            self.SendScintilla(self.SCI_SETINDICATORCURRENT, indicatorNumber);
+            self.SendScintilla(self.SCI_INDICATORFILLRANGE, start, finish - start);
+
+
 class SearchToolBar(QToolBar):
     conditionChanged = pyqtSignal(unicode, bool, bool)
     """Emitted (pattern, icase, wrap) when search condition changed"""
@@ -387,7 +444,7 @@ class SearchToolBar(QToolBar):
     searchRequested = pyqtSignal(unicode, bool, bool, bool)
     """Emitted (pattern, icase, wrap, forward) when requested"""
 
-    def __init__(self, parent=None, hidable=False, settings=None):
+    def __init__(self, parent=None, hidable=False):
         super(SearchToolBar, self).__init__(_('Search'), parent,
                                             objectName='search',
                                             iconSize=QSize(16, 16))
@@ -413,19 +470,22 @@ class SearchToolBar(QToolBar):
         self.addWidget(self._chk)
         self._wrapchk = QCheckBox(_('Wrap search'))
         self.addWidget(self._wrapchk)
-        self._bt = QPushButton(_('Search'), enabled=False)
+        self._btprev = QPushButton(_('Prev'), icon=qtlib.geticon('go-up'),
+                                   iconSize=QSize(16, 16))
+        self._btprev.clicked.connect(
+            lambda: self._emitSearchRequested(forward=False))
+        self.addWidget(self._btprev)
+        self._bt = QPushButton(_('Next'), icon=qtlib.geticon('go-down'),
+                               iconSize=QSize(16, 16))
         self._bt.clicked.connect(self._emitSearchRequested)
-        self._le.textChanged.connect(lambda s: self._bt.setEnabled(bool(s)))
+        self._le.textChanged.connect(self._updateSearchButtons)
         self.addWidget(self._bt)
 
         self.setFocusProxy(self._le)
         self.setStyleSheet(qtlib.tbstylesheet)
 
-        def defaultsettings():
-            s = QSettings()
-            s.beginGroup('searchtoolbar')
-            return s
-        self._settings = settings or defaultsettings()
+        self._settings = QSettings()
+        self._settings.beginGroup('searchtoolbar')
         self.searchRequested.connect(self._writesettings)
         self._readsettings()
 
@@ -433,6 +493,8 @@ class SearchToolBar(QToolBar):
         self._chk.toggled.connect(self._emitConditionChanged)
         self._wrapchk.toggled.connect(self._emitConditionChanged)
 
+        self._updateSearchButtons()
+
     def keyPressEvent(self, event):
         if event.matches(QKeySequence.FindNext):
             self._emitSearchRequested(forward=True)
@@ -478,6 +540,12 @@ class SearchToolBar(QToolBar):
         self.searchRequested.emit(self.pattern(), self.caseInsensitive(),
                                   self.wrapAround(), forward)
 
+    @pyqtSlot()
+    def _updateSearchButtons(self):
+        enabled = bool(self._le.text())
+        self._btprev.setEnabled(enabled)
+        self._bt.setEnabled(enabled)
+
     def pattern(self):
         """Returns the current search pattern [unicode]"""
         return self._le.text()
@@ -539,6 +607,15 @@ class KeyPressInterceptor(QObject):
             return True
         return False
 
+def unbindConflictedKeys(sci):
+    cmdset = sci.standardCommands()
+    try:
+        cmd = cmdset.boundTo(QKeySequence('CTRL+L'))
+        if cmd:
+            cmd.setKey(0)
+    except AttributeError:  # old QScintilla does not have boundTo()
+        pass
+
 def qsciEolModeFromOs():
     if os.name.startswith('nt'):
         return QsciScintilla.EolWindows
@@ -562,7 +639,7 @@ def findTabIndentsInLines(lines, linestocheck=100):
         elif line.startswith('\t'):
             return True
     return False # Use spaces for indents default
-    
+
 def fileEditor(filename, **opts):
     'Open a simple modal file editing dialog'
     dialog = QDialog()
@@ -586,7 +663,7 @@ def fileEditor(filename, **opts):
     def showsearchbar():
         searchbar.show()
         searchbar.setFocus(Qt.OtherFocusReason)
-    QShortcut(QKeySequence.Find, dialog, showsearchbar)
+    qtlib.newshortcutsforstdkey(QKeySequence.Find, dialog, showsearchbar)
     dialog.layout().addWidget(searchbar)
 
     BB = QDialogButtonBox
