@@ -15,6 +15,7 @@ import tempfile
 import re
 import weakref
 
+from mercurial.i18n import _ as hggettext
 from mercurial import commands, extensions, error, util
 
 from tortoisehg.util import hglib, paths
@@ -24,9 +25,8 @@ from hgext.color import _styles
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
-if PYQT_VERSION_STR.split('.') < ['4', '7'] or \
-   QT_VERSION_STR.split('.') < ['4', '6']:
-    sys.stderr.write('TortoiseHg requires Qt 4.6 and PyQt 4.7\n')
+if PYQT_VERSION < 0x40600 or QT_VERSION < 0x40600:
+    sys.stderr.write('TortoiseHg requires Qt 4.6 and PyQt 4.6\n')
     sys.stderr.write('You have Qt %s and PyQt %s\n' %
                      (QT_VERSION_STR, PYQT_VERSION_STR))
     sys.exit()
@@ -58,7 +58,7 @@ def gettempdir():
 def openhelpcontents(url):
     'Open online help, use local CHM file if available'
     if not url.startswith('http'):
-        fullurl = 'http://tortoisehg.org/manual/2.1/' + url
+        fullurl = 'http://tortoisehg.org/manual/2.7/' + url
         # Use local CHM file if it can be found
         if os.name == 'nt' and paths.bin_path:
             chm = os.path.join(paths.bin_path, 'doc', 'TortoiseHg.chm')
@@ -172,9 +172,9 @@ def savefiles(repo, files, rev, parent=None):
             filename = "%s@%d%s" % (wfile, rev, ext)
         else:
             filename = "%s@%d" % (ext, rev)
-        result = QFileDialog.getSaveFileName(parent=parent,
-                                             caption=_("Save file to"),
-                                             directory=filename) 
+        result = QFileDialog.getSaveFileName(
+            parent=parent, caption=_("Save file to"),
+            directory=hglib.tounicode(filename))
         if not result:
             continue
         cwd = os.getcwd()
@@ -184,7 +184,7 @@ def savefiles(repo, files, rev, parent=None):
                 commands.cat(repo.ui, repo, curfile, rev=rev,
                              output=hglib.fromunicode(result))
             except (util.Abort, IOError), e:
-                QMessageBox.critical(self, _('Unable to save file'),
+                QMessageBox.critical(parent, _('Unable to save file'),
                                      hglib.tounicode(str(e)))
         finally:
             os.chdir(cwd)
@@ -485,7 +485,7 @@ def geticon(name):
     This searches for the icon from theme, Qt resource or icons directory,
     named as 'name.(svg|png|ico)'.
     """
-    if QIcon.hasThemeIcon(name):
+    if hasattr(QIcon, 'hasThemeIcon') and QIcon.hasThemeIcon(name):
         return QIcon.fromTheme(name)
 
     try:
@@ -524,6 +524,19 @@ def getpixmap(name, width=16, height=16):
     _pixmapcache[key] = pixmap
     return pixmap
 
+def getcheckboxpixmap(state, bgcolor, widget):
+    pix = QPixmap(16,16)
+    painter = QPainter(pix)
+    painter.fillRect(0, 0, 16, 16, bgcolor)
+    option = QStyleOptionButton()
+    style = QApplication.style()
+    option.initFrom(widget)
+    option.rect = style.subElementRect(style.SE_CheckBoxIndicator, option)
+    option.rect.moveTo(1, 1)
+    option.state |= state
+    style.drawPrimitive(style.PE_IndicatorCheckBox, option, painter)
+    return pix
+
 class ThgFont(QObject):
     changed = pyqtSignal(QFont)
     def __init__(self, name):
@@ -693,6 +706,19 @@ def fix_application_font():
     f.setPixelSize(abs(lf.lfHeight))
     QApplication.setFont(f, 'QWidget')
 
+def allowCaseChangingInput(combo):
+    """Allow case-changing input of known combobox item
+
+    QComboBox performs case-insensitive inline completion by default. It's
+    all right, but sadly it implies case-insensitive check for duplicates,
+    i.e. you can no longer enter "Foo" if the combobox contains "foo".
+
+    For details, read QComboBoxPrivate::_q_editingFinished() and matchFlags()
+    of src/gui/widgets/qcombobox.cpp.
+    """
+    assert isinstance(combo, QComboBox) and combo.isEditable()
+    combo.completer().setCaseSensitivity(Qt.CaseSensitive)
+
 class PMButton(QPushButton):
     """Toggle button with plus/minus icon images"""
 
@@ -843,6 +869,57 @@ class LabeledSeparator(QWidget):
 
         self.setLayout(box)
 
+# Strings and regexes used to convert hashes and subrepo paths into links
+_hashregex = re.compile(r'\b([0-9a-fA-F]{12,})')
+# Currently converting subrepo paths into links only works in English
+_subrepoindicatorpattern = hglib.tounicode(hggettext('(in subrepo %s)') + '\n')
+
+def _linkifyHash(message, subrepo=''):
+    if subrepo:
+        p = 'repo:%s?' % subrepo
+    else:
+        p = 'cset:'
+    replaceexpr = lambda m: '<a href="%s">%s</a>' % (p + m.group(0), m.group(0))
+    return _hashregex.sub(replaceexpr, message)
+
+def _linkifySubrepoRef(message, subrepo, hash=''):
+    if hash:
+        hash = '?' + hash
+    subrepolink = '<a href="repo:%s%s">%s</a>' % (subrepo, hash, subrepo)
+    subrepoindicator = _subrepoindicatorpattern % subrepo
+    linkifiedsubrepoindicator = _subrepoindicatorpattern % subrepolink
+    message = message.replace(subrepoindicator, linkifiedsubrepoindicator)
+    return message
+
+def linkifyMessage(message, subrepo=None):
+    r"""Convert revision id hashes and subrepo paths in messages into links
+
+    >>> linkifyMessage('abort: 0123456789ab!\nhint: foo\n')
+    u'abort: <a href="cset:0123456789ab">0123456789ab</a>!<br>hint: foo<br>'
+    >>> linkifyMessage('abort: foo (in subrepo bar)\n', subrepo='bar')
+    u'abort: foo (in subrepo <a href="repo:bar">bar</a>)<br>'
+    >>> linkifyMessage('abort: 0123456789ab! (in subrepo bar)\nhint: foo\n',
+    ...                subrepo='bar') #doctest: +NORMALIZE_WHITESPACE
+    u'abort: <a href="repo:bar?0123456789ab">0123456789ab</a>!
+    (in subrepo <a href="repo:bar?0123456789ab">bar</a>)<br>hint: foo<br>'
+
+    subrepo name containing regexp backreference, \g:
+
+    >>> linkifyMessage('abort: 0123456789ab! (in subrepo foo\\goo)\n',
+    ...                subrepo='foo\\goo') #doctest: +NORMALIZE_WHITESPACE
+    u'abort: <a href="repo:foo\\goo?0123456789ab">0123456789ab</a>!
+    (in subrepo <a href="repo:foo\\goo?0123456789ab">foo\\goo</a>)<br>'
+    """
+    message = unicode(message)
+    message = _linkifyHash(message, subrepo)
+    if subrepo:
+        hash = ''
+        m = _hashregex.search(message)
+        if m:
+            hash = m.group(1)
+        message = _linkifySubrepoRef(message, subrepo, hash)
+    return message.replace('\n', '<br>')
+
 class InfoBar(QFrame):
     """Non-modal confirmation/alert (like web flash or Chrome's InfoBar)
 
@@ -903,8 +980,11 @@ class StatusInfoBar(InfoBar):
     """Show status message"""
     def __init__(self, message, parent=None):
         super(StatusInfoBar, self).__init__(parent)
-        self._msglabel = QLabel(message, self, wordWrap=True,
-                                textInteractionFlags=Qt.TextSelectableByMouse)
+        self._msglabel = QLabel(message, self,
+                                wordWrap=True,
+                                textInteractionFlags=Qt.TextSelectableByMouse \
+                                | Qt.LinksAccessibleByMouse)
+        self._msglabel.linkActivated.connect(self.linkActivated)
         self.addWidget(self._msglabel, stretch=1)
 
 class CommandErrorInfoBar(InfoBar):
@@ -914,8 +994,11 @@ class CommandErrorInfoBar(InfoBar):
     def __init__(self, message, parent=None):
         super(CommandErrorInfoBar, self).__init__(parent)
 
-        self._msglabel = QLabel(message, self, wordWrap=True,
-                                textInteractionFlags=Qt.TextSelectableByMouse)
+        self._msglabel = QLabel(message, self,
+                                wordWrap=True,
+                                textInteractionFlags=Qt.TextSelectableByMouse \
+                                | Qt.LinksAccessibleByMouse)
+        self._msglabel.linkActivated.connect(self.linkActivated)
         self.addWidget(self._msglabel, stretch=1)
 
         self._loglabel = QLabel('<a href="log:">%s</a>' % _('Show Log'))
@@ -934,7 +1017,9 @@ class ConfirmInfoBar(InfoBar):
         # no wordWrap=True and stretch=1, which inserts unwanted space
         # between _msglabel and _buttons.
         self._msglabel = QLabel(message, self,
-                                textInteractionFlags=Qt.TextSelectableByMouse)
+                                textInteractionFlags=Qt.TextSelectableByMouse \
+                                | Qt.LinksAccessibleByMouse)
+        self._msglabel.linkActivated.connect(self.linkActivated)
         self.addWidget(self._msglabel)
 
         self._buttons = QDialogButtonBox(self)
@@ -1076,6 +1161,17 @@ class Spacer(QWidget):
     def sizeHint(self):
         return QSize(self.width, self.height)
 
+def _configuredusername(ui):
+    # need to check the existence before calling ui.username(); otherwise it
+    # may fall back to the system default.
+    if (not os.environ.get('HGUSER') and not ui.config('ui', 'username')
+        and not os.environ.get('EMAIL')):
+        return None
+    try:
+        return ui.username()
+    except error.Abort:
+        return None
+
 def getCurrentUsername(widget, repo, opts=None):
     if opts:
         # 1. Override has highest priority
@@ -1084,10 +1180,9 @@ def getCurrentUsername(widget, repo, opts=None):
             return user
 
     # 2. Read from repository
-    try:
-        return repo.ui.username()
-    except error.Abort:
-        pass
+    user = _configuredusername(repo.ui)
+    if user:
+        return user
 
     # 3. Get a username from the user
     QMessageBox.information(widget, _('Please enter a username'),
@@ -1097,17 +1192,89 @@ def getCurrentUsername(widget, repo, opts=None):
     dlg = SettingsDialog(False, focus='ui.username')
     dlg.exec_()
     repo.invalidateui()
-    try:
-        return repo.ui.username()
-    except error.Abort:
-        return None
+    return _configuredusername(repo.ui)
+
+class _EncodingSafeInputDialog(QInputDialog):
+    def accept(self):
+        try:
+            hglib.fromunicode(self.textValue())
+            return super(_EncodingSafeInputDialog, self).accept()
+        except UnicodeEncodeError:
+            WarningMsgBox(_('Text Translation Failure'),
+                          _('Unable to translate input to local encoding.'),
+                          parent=self)
 
 def getTextInput(parent, title, label, mode=QLineEdit.Normal, text='',
-  flags=Qt.WindowFlags()):
-    # the flags argument is supported under Qt 4.6, but probably with
-    # a different name (see issue 252), so we simply call everything
-    # positionally
-    return QInputDialog.getText(parent, title, label, mode, text,
-      Qt.CustomizeWindowHint | Qt.WindowTitleHint |
-      Qt.WindowCloseButtonHint | flags)
+                 flags=Qt.WindowFlags()):
+    flags |= (Qt.CustomizeWindowHint | Qt.WindowTitleHint
+              | Qt.WindowCloseButtonHint)
+    dlg = _EncodingSafeInputDialog(parent, flags)
+    dlg.setWindowTitle(title)
+    dlg.setLabelText(label)
+    dlg.setTextValue(text)
+    dlg.setTextEchoMode(mode)
 
+    r = dlg.exec_()
+    dlg.setParent(None)  # so that garbage collected
+    return r and dlg.textValue() or '', bool(r)
+
+def keysequence(o):
+    """Create QKeySequence from string or QKeySequence"""
+    if isinstance(o, (QKeySequence, QKeySequence.StandardKey)):
+        return o
+    try:
+        return getattr(QKeySequence, str(o))  # standard key
+    except AttributeError:
+        return QKeySequence(o)
+
+def modifiedkeysequence(o, modifier):
+    """Create QKeySequence of modifier key prepended"""
+    origseq = QKeySequence(keysequence(o))
+    return QKeySequence('%s+%s' % (modifier, origseq.toString()))
+
+def newshortcutsforstdkey(key, *args, **kwargs):
+    """Create [QShortcut,...] for all key bindings of the given StandardKey"""
+    return [QShortcut(keyseq, *args, **kwargs)
+            for keyseq in QKeySequence.keyBindings(key)]
+
+class PaletteSwitcher(object):
+    """
+    Class that can be used to enable a predefined, alterantive background color
+    for a widget
+
+    This is normally used to change the color of widgets when they display some
+    "filtered" content which is a subset of the actual widget contents.
+
+    The alternative background color is fixed, and depends on the original
+    background color (dark and light backgrounds use a different alternative
+    color).
+
+    The alterenative color cannot be changed because the idea is to set a
+    consistent "filter" style for all widgets.
+
+    An instance of this class must be added as a property of the widget whose
+    background we want to change. The constructor takes the "target widget" as
+    its only parameter.
+
+    In order to enable or disable the background change, simply call the
+    enablefilterpalette() method.
+    """
+    def __init__(self, targetwidget):
+        self._targetwidget = targetwidget
+        self._defaultpalette = targetwidget.palette()
+        bgcolor = self._defaultpalette.color(QPalette.Base)
+        if bgcolor.black() <= 128:
+            # Light theme
+            filterbgcolor = QColor('#FFFFB7')
+        else:
+            # Dark theme
+            filterbgcolor = QColor('darkgrey')
+        self._filterpalette = QPalette()
+        self._filterpalette.setColor(QPalette.Base, filterbgcolor)
+
+    def enablefilterpalette(self, enabled=False):
+        if enabled:
+            pl = self._filterpalette
+        else:
+            pl = self._defaultpalette
+        self._targetwidget.setPalette(pl)
