@@ -7,7 +7,7 @@
 
 import os
 
-from mercurial import commands, error, hg, ui, util
+from mercurial import commands, hg, ui, util
 
 from tortoisehg.util import hglib, paths
 from tortoisehg.hgqt.i18n import _
@@ -45,7 +45,8 @@ class RepoTreeView(QTreeView):
         self.setAcceptDrops(True)
         self.setAutoScroll(True)
         self.setDragDropMode(QAbstractItemView.DragDrop)
-        self.setDefaultDropAction(Qt.MoveAction)
+        if PYQT_VERSION >= 0x40700:
+            self.setDefaultDropAction(Qt.MoveAction)
         self.setDropIndicatorShown(True)
         self.setEditTriggers(QAbstractItemView.DoubleClicked
                              | QAbstractItemView.EditKeyPressed)
@@ -120,21 +121,21 @@ class RepoTreeView(QTreeView):
         index, group, row = self.dropLocation(event)
 
         if index:
+            m = self.model()
             if event.source() is self:
                 # Event is an internal move, so pass it to the model
                 col = 0
-                drop = self.model().dropMimeData(data, event.dropAction(), row,
-                                                 col, group)
-                if drop:
+                if m.dropMimeData(data, event.dropAction(), row, col, group):
                     event.accept()
                     self.dropAccepted.emit()
             else:
                 # Event is a drop of an external repo
                 accept = False
                 for u in data.urls():
-                    root = paths.find_root(hglib.fromunicode(u.toLocalFile()))
-                    if root and not self.model().getRepoItem(root):
-                        self.model().addRepo(group, root, row)
+                    uroot = paths.find_root(unicode(u.toLocalFile()))
+                    if uroot and not m.isKnownRepoRoot(uroot, standalone=True):
+                        repoindex = m.addRepo(uroot, row, group)
+                        m.loadSubrepos(repoindex)
                         accept = True
                 if accept:
                     event.setDropAction(Qt.LinkAction)
@@ -200,6 +201,8 @@ class RepoTreeView(QTreeView):
     def removeSelected(self):
         'remove selected repository'
         s = self.selitem
+        if not s:
+            return
         item = s.internalPointer()
         if 'remove' not in item.menulist():  # check capability
             return
@@ -224,14 +227,11 @@ class RepoRegistryView(QDockWidget):
     removeRepo = pyqtSignal(QString)
     progressReceived = pyqtSignal(QString, object, QString, QString, object)
 
-    def __init__(self, parent, showSubrepos=False, showNetworkSubrepos=False,
-            showShortPaths=False):
+    def __init__(self, parent):
         QDockWidget.__init__(self, parent)
 
         self.watcher = None
-        self.showSubrepos = showSubrepos
-        self.showNetworkSubrepos = showNetworkSubrepos
-        self.showShortPaths = showShortPaths
+        self._setupSettingActions()
 
         self.setFeatures(QDockWidget.DockWidgetClosable |
                          QDockWidget.DockWidgetMovable  |
@@ -245,12 +245,6 @@ class RepoRegistryView(QDockWidget):
 
         self.contextmenu = QMenu(self)
         self.tview = tv = RepoTreeView(self)
-
-        sfile = settingsfilename()
-        tv.setModel(repotreemodel.RepoTreeModel(sfile, self,
-            showSubrepos=self.showSubrepos,
-            showNetworkSubrepos=self.showNetworkSubrepos))
-
         mainframe.layout().addWidget(tv)
 
         tv.setIndentation(10)
@@ -264,7 +258,13 @@ class RepoRegistryView(QDockWidget):
         tv.dropAccepted.connect(self.dropAccepted)
 
         self.createActions()
-        QTimer.singleShot(0, self.expand)
+        self._loadSettings()
+        self._updateSettingActions()
+
+        sfile = settingsfilename()
+        model = repotreemodel.RepoTreeModel(sfile, self,
+            showShortPaths=self._isSettingEnabled('showShortPaths'))
+        tv.setModel(model)
 
         # Setup a file system watcher to update the reporegistry
         # anytime it is modified by another thg instance
@@ -276,28 +276,70 @@ class RepoRegistryView(QDockWidget):
             tv.model().write(sfile)
         self.watcher = QFileSystemWatcher(self)
         self.watcher.addPath(sfile)
-        self.watcher.fileChanged.connect(self.modifiedSettings)
-        self._pendingReloadModel = False
-        self._activeTabRepo = None
+        self._reloadModelTimer = QTimer(self, interval=2000, singleShot=True)
+        self._reloadModelTimer.timeout.connect(self.reloadModel)
+        self.watcher.fileChanged.connect(self._reloadModelTimer.start)
 
-    def setShowSubrepos(self, show, reloadModel=True):
-        if self.showSubrepos != show:
-            self.showSubrepos = show
-            if reloadModel:
-                self.reloadModel()
+        QTimer.singleShot(0, self._initView)
 
-    def setShowNetworkSubrepos(self, show, reloadModel=True):
-        if self.showNetworkSubrepos != show:
-            self.showNetworkSubrepos = show
-            if reloadModel:
-                self.reloadModel()
+    @pyqtSlot()
+    def _initView(self):
+        self.expand()
+        self._updateColumnVisibility()
+        if self._isSettingEnabled('showSubrepos'):
+            self._scanAllRepos()
 
-    def setShowShortPaths(self, show):
-        if self.showShortPaths != show:
-            self.showShortPaths = show
-            #self.tview.model().showShortPaths = show
-            self.tview.model().updateCommonPaths(show)
-            self.tview.dataChanged(QModelIndex(), QModelIndex())
+    def _loadSettings(self):
+        defaultmap = {'showPaths': False, 'showSubrepos': False,
+                      'showNetworkSubrepos': False, 'showShortPaths': True}
+        s = QSettings()
+        s.beginGroup('Workbench')  # for compatibility with old release
+        for key, action in self._settingactions.iteritems():
+            action.setChecked(s.value(key, defaultmap[key]).toBool())
+        s.endGroup()
+
+    def _saveSettings(self):
+        s = QSettings()
+        s.beginGroup('Workbench')  # for compatibility with old release
+        for key, action in self._settingactions.iteritems():
+            s.setValue(key, action.isChecked())
+        s.endGroup()
+
+    def _setupSettingActions(self):
+        settingtable = [
+            ('showPaths', _('Show &Paths'), self._updateColumnVisibility),
+            ('showShortPaths', _('Show S&hort Paths'), self._updateCommonPath),
+            ('showSubrepos', _('&Scan Repositories at Startup'), None),
+            ('showNetworkSubrepos', _('Scan &Remote Repositories'), None),
+            ]
+        self._settingactions = {}
+        for i, (key, text, slot) in enumerate(settingtable):
+            a = QAction(text, self, checkable=True)
+            a.setData(i)  # sort key
+            if slot:
+                a.triggered.connect(slot)
+            a.triggered.connect(self._updateSettingActions)
+            self._settingactions[key] = a
+
+    @pyqtSlot()
+    def _updateSettingActions(self):
+        ax = self._settingactions
+        ax['showNetworkSubrepos'].setEnabled(ax['showSubrepos'].isChecked())
+        ax['showShortPaths'].setEnabled(ax['showPaths'].isChecked())
+
+    def settingActions(self):
+        return sorted(self._settingactions.itervalues(),
+                      key=lambda a: a.data().toInt())
+
+    def _isSettingEnabled(self, key):
+        return self._settingactions[key].isChecked()
+
+    @pyqtSlot()
+    def _updateCommonPath(self):
+        show = self._isSettingEnabled('showShortPaths')
+        self.tview.model().updateCommonPaths(show)
+        # FIXME: access violation; should be done by model
+        self.tview.dataChanged(QModelIndex(), QModelIndex())
 
     def updateSettingsFile(self):
         # If there is a settings watcher, we must briefly stop watching the
@@ -320,87 +362,45 @@ class RepoRegistryView(QDockWidget):
         # file
         QTimer.singleShot(0, self.updateSettingsFile)
 
-    @pyqtSlot(QString)
-    def modifiedSettings(self):
-        UPDATE_DELAY = 2 # seconds
-
-        # Do not update the repo registry more often than
-        # once every UPDATE_DELAY seconds
-        if not self._pendingReloadModel:
-            # There are no pending updates:
-            # -> schedule and update in UPDATE_DELAY seconds.
-            # If other update notifications arrive from now
-            # until now + UPDATE_DELAY, they will be ignored and "rolled into"
-            # the pending update
-            self._pendingReloadModel = True
-            QTimer.singleShot(1000 * UPDATE_DELAY, self.reloadModel)
-
+    @pyqtSlot()
     def reloadModel(self):
-        self.tview.setModel(
-            repotreemodel.RepoTreeModel(settingsfilename(), self,
-                self.showSubrepos, self.showNetworkSubrepos,
-                self.showShortPaths))
+        oldmodel = self.tview.model()
+        activeroot = oldmodel.repoRoot(oldmodel.activeRepoIndex())
+        newmodel = repotreemodel.RepoTreeModel(settingsfilename(), self,
+            self._isSettingEnabled('showShortPaths'))
+        self.tview.setModel(newmodel)
+        oldmodel.deleteLater()
+        if self._isSettingEnabled('showSubrepos'):
+            self._scanAllRepos()
         self.expand()
-        self._pendingReloadModel = False
+        if activeroot:
+            self.setActiveTabRepo(activeroot)
+        self._reloadModelTimer.stop()
 
-    def expand(self, it=None):
-        if not it:
-            self.tview.expandToDepth(0)
-        else:
-            # Create a list of ancestors (including the selected item)
-            from repotreeitem import RepoGroupItem
-            itchain = [it]
-            while(not isinstance(itchain[-1], RepoGroupItem)):
-                itchain.append(itchain[-1].parent())
+    def expand(self):
+        self.tview.expandToDepth(0)
 
-            # Starting from the topmost ancestor (a root item), expand the
-            # ancestors one by one
-            m = self.tview.model()
-            idx = self.tview.rootIndex()
-            for it in reversed(itchain):
-                idx = m.index(it.row(), 0, idx)
-                self.tview.expand(idx)
-
-    def addRepo(self, root, groupname=None):
-        """
-        Add a repo to the repo registry, optionally specifying the parent repository group
-
-        The main use of this method is when the workbench has opened a new repowidget
-        """
+    def addRepo(self, uroot):
+        """Add repo if not exists; called when the workbench has opened it"""
         m = self.tview.model()
-        it = m.getRepoItem(root, lookForSubrepos=True)
-        if it == None:
-            group = None
-            if groupname:
-                # Get the group index of the RepoGroup corresponding to the target group name
-                for it in m.rootItem.childs:
-                    if groupname == it.name:
-                        rootidx = self.tview.rootIndex()
-                        group = m.index(it.row(), 0, rootidx)
-                        break
-            m.addRepo(group, root, -1)
+        knownindex = m.indexFromRepoRoot(uroot)
+        if knownindex.isValid():
+            self._scanAddedRepo(knownindex)  # just scan stale subrepos
+        else:
+            index = m.addRepo(uroot)
+            self._scanAddedRepo(index)
             self.updateSettingsFile()
 
     def setActiveTabRepo(self, root):
-        """"
-        The selected tab has changed on the workbench
-        Unmark the previously selected tab and mark the new one as selected on
-        the Repo Registry as well
-        """
-        root = hglib.fromunicode(root)
-        if self._activeTabRepo:
-            self._activeTabRepo.setActive(False)
+        """"The selected tab has changed on the workbench"""
         m = self.tview.model()
-        it = m.getRepoItem(root, lookForSubrepos=True)
-        if it:
-            self._activeTabRepo = it
-            it.setActive(True)
-            self.tview.dataChanged(QModelIndex(), QModelIndex())
+        index = m.indexFromRepoRoot(root)
+        m.setActiveRepo(index)
+        self.tview.scrollTo(index)
 
-            # Make sure that the active tab is visible by expanding its parent
-            self.expand(it.parent())
-
-    def showPaths(self, show):
+    @pyqtSlot()
+    def _updateColumnVisibility(self):
+        show = self._isSettingEnabled('showPaths')
         self.tview.setColumnHidden(1, not show)
         self.tview.setHeaderHidden(not show)
         if show:
@@ -412,43 +412,44 @@ class RepoRegistryView(QDockWidget):
         sfile = settingsfilename()
         self.watcher.removePath(sfile)
         self.tview.model().write(sfile)
+        self._saveSettings()
 
     def _action_defs(self):
-        a = [("reloadRegistry", _("Refresh repository list"), 'view-refresh',
+        a = [("reloadRegistry", _("&Refresh Repository List"), 'view-refresh',
                 _("Refresh the Repository Registry list"), self.reloadModel),
-             ("open", _("Open"), 'thg-repository-open',
+             ("open", _("&Open"), 'thg-repository-open',
                 _("Open the repository in a new tab"), self.open),
-             ("openAll", _("Open All"), 'thg-repository-open',
+             ("openAll", _("&Open All"), 'thg-repository-open',
                 _("Open all repositories in new tabs"), self.openAll),
-             ("newGroup", _("New Group"), 'new-group',
+             ("newGroup", _("New &Group"), 'new-group',
                 _("Create a new group"), self.newGroup),
-             ("rename", _("Rename"), None,
+             ("rename", _("Re&name"), None,
                 _("Rename the entry"), self.startRename),
-             ("settings", _("Settings..."), 'settings_user',
+             ("settings", _("Settin&gs"), 'settings_user',
                 _("View the repository's settings"), self.startSettings),
-             ("remove", _("Remove from registry"), 'menudelete',
+             ("remove", _("Re&move from Registry"), 'menudelete',
                 _("Remove the node and all its subnodes."
                   " Repositories are not deleted from disk."),
                   self.removeSelected),
-             ("clone", _("Clone..."), 'hg-clone',
+             ("clone", _("Clon&e..."), 'hg-clone',
                 _("Clone Repository"), self.cloneRepo),
-             ("explore", _("Explore"), 'system-file-manager',
+             ("explore", _("E&xplore"), 'system-file-manager',
                 _("Open the repository in a file browser"), self.explore),
-             ("terminal", _("Terminal"), 'utilities-terminal',
+             ("terminal", _("&Terminal"), 'utilities-terminal',
                 _("Open a shell terminal in the repository root"), self.terminal),
-             ("add", _("Add repository..."), 'hg',
+             ("add", _("&Add Repository..."), 'hg',
                 _("Add a repository to this group"), self.addNewRepo),
-             ("addsubrepo", _("Add a subrepository..."), 'thg-add-subrepo',
+             ("addsubrepo", _("A&dd Subrepository..."), 'thg-add-subrepo',
                 _("Convert an existing repository into a subrepository"),
                 self.addSubrepo),
-             ("copypath", _("Copy path"), '',
+             ("copypath", _("Copy &Path"), '',
                 _("Copy the root path of the repository to the clipboard"),
                 self.copyPath),
-             ("sortbyname", _("Sort by name"), '',
+             ("sortbyname", _("Sort by &Name"), '',
                 _("Sort the group by short name"), self.sortbyname),
-             ("sortbypath", _("Sort by path"), '',
+             ("sortbypath", _("Sort by &Path"), '',
                 _("Sort the group by full path"), self.sortbypath),
-             ("sortbyhgsub", _("Sort by .hgsub"), '',
+             ("sortbyhgsub", _("&Sort by .hgsub"), '',
                 _("Order the subrepos as in .hgsub"), self.sortbyhgsub),
              ]
         return a
@@ -497,7 +498,7 @@ class RepoRegistryView(QDockWidget):
         root = self.selitem.internalPointer().rootpath()
         d = clone.CloneDialog(args=[root, root + '-clone'], parent=self)
         d.finished.connect(d.deleteLater)
-        d.clonedRepository.connect(self.openClone)
+        d.clonedRepository.connect(self._openClone)
         d.show()
 
     def explore(self):
@@ -515,15 +516,11 @@ class RepoRegistryView(QDockWidget):
         path = FD.getExistingDirectory(caption=caption,
                                        options=FD.ShowDirsOnly | FD.ReadOnly)
         if path:
-            root = paths.find_root(hglib.fromunicode(path))
-            if root and not self.tview.model().getRepoItem(root):
-                try:
-                    self.tview.model().addRepo(self.selitem, root)
-                except error.RepoError:
-                    qtlib.WarningMsgBox(
-                        _('Failed to add repository'),
-                        _('%s is not a valid repository') % path, parent=self)
-                    return
+            m = self.tview.model()
+            uroot = paths.find_root(unicode(path))
+            if uroot and not m.isKnownRepoRoot(uroot, standalone=True):
+                index = m.addRepo(uroot, parent=self.selitem)
+                self._scanAddedRepo(index)
 
     def addSubrepo(self):
         'menu action handler for adding a new subrepository'
@@ -533,7 +530,7 @@ class RepoRegistryView(QDockWidget):
         path = unicode(FD.getExistingDirectory(caption=caption,
             directory=root, options=FD.ShowDirsOnly | FD.ReadOnly))
         if path:
-            path = os.path.normcase(os.path.normpath(path))
+            path = os.path.normpath(path)
             sroot = paths.find_root(path)
 
             root = os.path.normcase(os.path.normpath(root))
@@ -548,12 +545,12 @@ class RepoRegistryView(QDockWidget):
                     _('"%s" is not a folder') % sroot,
                     parent=self)
                 return
-            elif sroot == root:
+            elif os.path.normcase(sroot) == root:
                 qtlib.WarningMsgBox(_('Cannot add subrepository'),
                     _('A repository cannot be added as a subrepo of itself'),
                     parent=self)
                 return
-            elif root != paths.find_root(os.path.dirname(path)):
+            elif root != paths.find_root(os.path.dirname(os.path.normcase(path))):
                 qtlib.WarningMsgBox(_('Cannot add subrepository'),
                     _('The selected folder:<br><br>%s<br><br>'
                     'is not inside the target repository.<br><br>'
@@ -619,6 +616,8 @@ class RepoRegistryView(QDockWidget):
                                 _('The .hgsub file already contains the '
                                 'line:<br><br>%s') % line, parent=self)
                             return
+                    if not linesep:
+                        linesep = os.linesep
 
                     # Append the new subrepo to the end of the .hgsub file
                     lines.append(hglib.fromunicode('%s = %s'
@@ -628,7 +627,7 @@ class RepoRegistryView(QDockWidget):
                     # and update the .hgsub file
                     try:
                         fsub = repo.wopener('.hgsub', 'w')
-                        fsub.write(linesep.join(lines))
+                        fsub.write(linesep.join(lines) + linesep)
                         fsub.close()
                         if not hasHgsub:
                             commands.add(ui.ui(), repo, repo.wjoin('.hgsub'))
@@ -665,16 +664,16 @@ class RepoRegistryView(QDockWidget):
         for root in self.selitem.internalPointer().childRoots():
             self.openRepo.emit(hglib.tounicode(root), False)
 
-    def openClone(self, root=None, sourceroot=None):
+    @pyqtSlot(unicode, unicode)
+    def _openClone(self, root, sourceroot):
         m = self.tview.model()
-        src = m.getRepoItem(hglib.fromunicode(sourceroot))
-        if src:
-            groupname = src.parent().name
-        else:
-            groupname = None
-        self.open(root, groupname)
+        src = m.indexFromRepoRoot(sourceroot, standalone=True)
+        if src.isValid() and not m.isKnownRepoRoot(root):
+            index = m.addRepo(root, parent=src.parent())
+            self._scanAddedRepo(index)
+        self.open(root)
 
-    def open(self, root=None, groupname=None):
+    def open(self, root=None):
         'open context menu action, open repowidget unconditionally'
         if not root:
             root = self.selitem.internalPointer().rootpath()
@@ -686,8 +685,6 @@ class RepoRegistryView(QDockWidget):
             else:
                 repotype = 'unknown'
         if repotype == 'hg':
-            if groupname:
-                self.addRepo(hglib.tounicode(root), groupname)
             self.openRepo.emit(hglib.tounicode(root), False)
         else:
             qtlib.WarningMsgBox(
@@ -719,7 +716,7 @@ class RepoRegistryView(QDockWidget):
 
     def sortbyname(self):
         childs = self.selitem.internalPointer().childs
-        self.tview.model().sortchilds(childs, lambda x: x.shortname())
+        self.tview.model().sortchilds(childs, lambda x: x.shortname().lower())
 
     def sortbypath(self):
         childs = self.selitem.internalPointer().childs
@@ -758,24 +755,45 @@ class RepoRegistryView(QDockWidget):
         if it:
             it.setBaseNode(basenode)
 
-    @pyqtSlot(QString)
-    def repoChanged(self, uroot):
+    def _scanAddedRepo(self, index):
         m = self.tview.model()
-        changedrootpath = hglib.fromunicode(QDir.fromNativeSeparators(uroot))
+        invalidpaths = m.loadSubrepos(index)
+        if not invalidpaths:
+            return
 
-        def isAboveOrBelowUroot(testedpath):
-            """Return True if rootpath is contained or contains uroot"""
-            r1 = hglib.fromunicode(QDir.fromNativeSeparators(testedpath)) + "/"
-            r2 = changedrootpath + "/"
-            return r1.startswith(r2) or r2.startswith(r1)
-
-        m.loadSubrepos(m.rootItem, isAboveOrBelowUroot)
-
-    @pyqtSlot(int, int, QString, QString)
-    def updateProgress(self, pos, max, topic, item):
-        if pos == max:
-            #self.progressReceived.emit('Updating repository registry', None, '', '', None)
-            self.progressReceived.emit(topic, None, item, '', None)
+        root = m.repoRoot(index)
+        if root in invalidpaths:
+            qtlib.WarningMsgBox(_('Could not get subrepository list'),
+                _('It was not possible to get the subrepository list for '
+                  'the repository in:<br><br><i>%s</i>') % root, parent=self)
         else:
-            #self.progressReceived.emit('Updating repository registry', pos, 'reporegistry-%s' % topic, '', max)
-            self.progressReceived.emit(topic, pos, item, '', max)
+            qtlib.WarningMsgBox(_('Could not open some subrepositories'),
+                _('It was not possible to fully load the subrepository '
+                  'list for the repository in:<br><br><i>%s</i><br><br>'
+                  'The following subrepositories may be missing, broken or '
+                  'on an inconsistent state and cannot be accessed:'
+                  '<br><br><i>%s</i>')
+                % (root, "<br>".join(invalidpaths)), parent=self)
+
+    @pyqtSlot(QString)
+    def scanRepo(self, uroot):
+        m = self.tview.model()
+        index = m.indexFromRepoRoot(uroot)
+        if index.isValid():
+            m.loadSubrepos(index)
+
+    def _scanAllRepos(self):
+        m = self.tview.model()
+        indexes = m.indexesOfRepoItems(standalone=True)
+        if not self._isSettingEnabled('showNetworkSubrepos'):
+            indexes = [idx for idx in indexes
+                       if not paths.netdrive_status(m.repoRoot(idx))]
+
+        topic = _('Updating repository registry')
+        for n, idx in enumerate(indexes):
+            self.progressReceived.emit(
+                topic, n, _('Loading repository %s') % m.repoRoot(idx), '',
+                len(indexes))
+            m.loadSubrepos(idx)
+        self.progressReceived.emit(
+            topic, None, _('Repository Registry updated'), '', None)
