@@ -8,12 +8,9 @@
 
 import binascii
 import os
-import shlex, subprocess, functools # used by runCustomCommand
-from mercurial import revset, error, patch, commands
-
-# hg >= 2.1
-if hasattr(commands, 'phase'):
-    from mercurial import phases
+import shlex, subprocess  # used by runCustomCommand
+import urllib
+from mercurial import revset, error, patch, phases
 
 from tortoisehg.util import hglib, shlib, paths
 from tortoisehg.hgqt.i18n import _
@@ -25,7 +22,7 @@ from tortoisehg.hgqt import cmdui, update, tag, backout, merge, visdiff
 from tortoisehg.hgqt import archive, thgimport, thgstrip, run, purge, bookmark
 from tortoisehg.hgqt import bisect, rebase, resolve, thgrepo, compress, mq
 from tortoisehg.hgqt import qdelete, qreorder, qfold, qrename, shelve
-from tortoisehg.hgqt import matching
+from tortoisehg.hgqt import matching, graft
 
 from tortoisehg.hgqt.repofilter import RepoFilterBar
 from tortoisehg.hgqt.repoview import HgRepoView
@@ -39,8 +36,6 @@ from tortoisehg.hgqt.pbranch import PatchBranchWidget
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
-import functools
-
 class RepoWidget(QWidget):
 
     showMessageSignal = pyqtSignal(QString)
@@ -76,7 +71,6 @@ class RepoWidget(QWidget):
         repo.repositoryDestroyed.connect(self.repositoryDestroyed)
         repo.configChanged.connect(self.configChanged)
         self.revsetfilter = False
-        self.ubranch = u''
         self.bundle = None  # bundle file name [local encoding]
         self.bundlesource = None  # source URL of incoming bundle [unicode]
         self.outgoingMode = False
@@ -175,7 +169,7 @@ class RepoWidget(QWidget):
         self.filterbar.progress.connect(self.progress)
         self.filterbar.showMessage.connect(self.showMessage)
         self.filterbar.setRevisionSet.connect(self.setRevisionSet)
-        self.filterbar.clearRevisionSet.connect(self.clearRevisionSet)
+        self.filterbar.clearRevisionSet.connect(self._unapplyRevisionSet)
         self.filterbar.filterToggled.connect(self.filterToggled)
         self.filterbar.hide()
         self.revsetfilter = self.filterbar.filtercb.isChecked()
@@ -255,12 +249,17 @@ class RepoWidget(QWidget):
         else:
             self.pbranchTabIndex = -1
 
+        self.taskTabsWidget.currentChanged.connect(
+            self._refreshCommitTabIfNeeded)
+
     @pyqtSlot(QString)
     def switchToNamedTaskTab(self, tabname):
         tabname = str(tabname)
         if tabname in self.namedTabs:
             idx = self.namedTabs[tabname]
-            if tabname == 'commit':
+            # refresh status even if current widget is already a 'commit'
+            if (tabname == 'commit'
+                and self.taskTabsWidget.currentIndex() == idx):
                 self._refreshCommitTabIfNeeded()
             self.taskTabsWidget.setCurrentIndex(idx)
 
@@ -272,8 +271,8 @@ class RepoWidget(QWidget):
         """Returns the expected title for this widget [unicode]"""
         if self.bundle:
             return _('%s <incoming>') % self.repo.shortname
-        elif self.ubranch:
-            return u'%s [%s]' % (self.repo.shortname, self.ubranch)
+        elif self.repomodel.branch():
+            return u'%s [%s]' % (self.repo.shortname, self.repomodel.branch())
         else:
             return self.repo.shortname
 
@@ -285,20 +284,28 @@ class RepoWidget(QWidget):
         """Toggle display repowidget filter bar"""
         self.filterbar.setVisible(checked)
 
+    def _openRepoLink(self, upath):
+        path = hglib.fromunicode(upath)
+        if not os.path.isabs(path):
+            path = self.repo.wjoin(path)
+        self.repoLinkClicked.emit(hglib.tounicode(path))
+
     @pyqtSlot(unicode)
     def _openLink(self, link):
         link = unicode(link)
         handlers = {'cset': self.goto,
                     'log': lambda a: self.makeLogVisible.emit(True),
-                    'subrepo': self.repoLinkClicked.emit,
+                    'repo': self._openRepoLink,
                     'shelve' : self.shelve}
         if ':' in link:
             scheme, param = link.split(':', 1)
             hdr = handlers.get(scheme)
             if hdr:
                 return hdr(param)
-
-        QDesktopServices.openUrl(QUrl(link))
+        if os.path.isabs(link):
+            qtlib.openlocalurl(link)
+        else:
+            QDesktopServices.openUrl(QUrl(link))
 
     def setInfoBar(self, cls, *args, **kwargs):
         """Show the given infobar at top of RepoWidget
@@ -334,6 +341,15 @@ class RepoWidget(QWidget):
     def _showOutputOnInfoBar(self, msg, label, maxlines=2, maxwidth=140):
         labelslist = unicode(label).split()
         if 'ui.error' in labelslist:
+            # Check if a subrepo is set in the label list
+            subrepo = None
+            subrepolabel = 'subrepo='
+            for label in labelslist:
+                if label.startswith(subrepolabel):
+                    # The subrepo "label" is encoded ascii
+                    subrepo = hglib.tounicode(
+                        urllib.unquote(str(label)[len(subrepolabel):]))
+                    break
             # Limit the text shown on the info bar to maxlines lines of up to maxwidth chars
             msglines = unicode(msg).strip().splitlines()
             infolines = []
@@ -343,7 +359,8 @@ class RepoWidget(QWidget):
                 infolines.append(line)
             if len(msglines) > maxlines and not infolines[-1].endswith('...'):
                 infolines[-1] += ' ...'
-            self.setInfoBar(qtlib.CommandErrorInfoBar, '\n'.join(infolines))
+            infomsg = qtlib.linkifyMessage('\n'.join(infolines), subrepo=subrepo)
+            self.setInfoBar(qtlib.CommandErrorInfoBar, infomsg)
 
     @pyqtSlot(unicode)
     def _showMessageOnInfoBar(self, msg):
@@ -355,18 +372,7 @@ class RepoWidget(QWidget):
     def createCommitWidget(self):
         pats, opts = {}, {}
         cw = CommitWidget(self.repo, pats, opts, True, self, rev=self.rev)
-
-        if cw.hasmqbutton:
-            cw.buttonHBox.addWidget(cw.mqSetupButton())
-        else:
-            b = QPushButton(_('Commit', 'action button'))
-            b.setAutoDefault(True)
-            f = b.font()
-            f.setWeight(QFont.Bold)
-            b.setFont(f)
-            cw.buttonHBox.addWidget(b)
-            cw.commitButtonEnable.connect(b.setEnabled)
-            b.clicked.connect(cw.commit)
+        cw.buttonHBox.addWidget(cw.commitSetupButton())
         cw.loadSettings(QSettings(), 'workbench')
 
         cw.output.connect(self.output)
@@ -390,6 +396,7 @@ class RepoWidget(QWidget):
         w.linkActivated.connect(self._openLink)
         w.showMessage.connect(self.showMessage)
         w.grepRequested.connect(self.grep)
+        w.revsetFilterRequested.connect(self.setFilter)
         return w
 
     def createSyncWidget(self):
@@ -426,7 +433,7 @@ class RepoWidget(QWidget):
         else:
             self.showIcon.emit(QIcon())
 
-    @pyqtSlot(QString, QString)
+    @pyqtSlot(QString)
     def setFilter(self, filter):
         self.filterbar.revsetle.setText(filter)
         self.filterbar.setVisible(True)
@@ -518,8 +525,13 @@ class RepoWidget(QWidget):
     @pyqtSlot()
     def clearRevisionSet(self):
         self.filterbar.revsetle.clear()
+        return self._unapplyRevisionSet()
+
+    @pyqtSlot()
+    def _unapplyRevisionSet(self):
         self.toolbarVisibilityChanged.emit()
         self.outgoingMode = False
+        self.repoview.enablefilterpalette(False)
         if not self.revset:
             return False
         self.revset = []
@@ -542,6 +554,7 @@ class RepoWidget(QWidget):
             self.refresh()
         self.repoview.resetBrowseHistory(self.revset)
         self._reload_rev = self.revset[0]
+        self.repoview._paletteswitcher.enablefilterpalette(revs)
 
     @pyqtSlot(bool)
     def filterToggled(self, checked):
@@ -554,22 +567,32 @@ class RepoWidget(QWidget):
     def setOutgoingNodes(self, nodes):
         self.filterbar.revsetle.setText('outgoing()')
         self.setRevisionSet([self.repo[n].rev() for n in nodes])
+        self.outgoingMode = True
 
         w = self.setInfoBar(qtlib.ConfirmInfoBar,
                             _('%d outgoing changesets') % len(nodes))
         assert w
 
         # Read the tortoisehg.defaultpush setting to determine what to push
-        # by default
+        # by default, and set the button label and action accordingly
         acceptbuttontext = _('Push')
         defaultpush = self.repo.ui.config('tortoisehg', 'defaultpush', 'all')
+        rev = None
+        branch = None
+        pushall = False
         if defaultpush == 'branch':
-            acceptbuttontext = _('Push current branch')
+            branch = self.repo['.'].branch()
+            acceptbuttontext = _('Push current branch (%s)') \
+                % hglib.tounicode(branch)
         elif defaultpush == 'revision':
-            acceptbuttontext = _('Push current revision')
+            rev = self.repo['.'].rev()
+            acceptbuttontext = _('Push current revision (%d)') % rev
+        else:
+            pushall = True
 
         w.acceptButton.setText(acceptbuttontext)
-        w.accepted.connect(lambda: self.push(False))  # TODO: to the same URL
+        w.accepted.connect(lambda: self.push(False,
+            rev=rev, branch=branch, pushall=pushall))  # TODO: to the same URL
         w.rejected.connect(self.clearRevisionSet)
 
     def createGrepWidget(self):
@@ -718,8 +741,6 @@ class RepoWidget(QWidget):
             if os.path.exists(self.repo.sjoin('undo')):
                 try:
                     args = self.repo.opener('undo.desc', 'r').read().splitlines()
-                    if args[1] != 'commit':
-                        return None
                     return args[1], int(args[0])
                 except (IOError, IndexError, ValueError):
                     pass
@@ -777,12 +798,15 @@ class RepoWidget(QWidget):
         # Filter revision set in case revisions were removed
         self.revset = [r for r in self.revset if r < len(self.repo)]
         self.repomodel = HgRepoListModel(self.repo, self.repoview.colselect[0],
-                                         self.ubranch, self.revset,
+                                         self.filterbar.branch(), self.revset,
                                          self.revsetfilter, self)
         self.repomodel.filled.connect(self.modelFilled)
         self.repomodel.loaded.connect(self.modelLoaded)
         self.repomodel.showMessage.connect(self.showMessage)
+        oldmodel = self.repoview.model()
         self.repoview.setModel(self.repomodel)
+        if oldmodel:
+            oldmodel.deleteLater()
         try:
             self._last_series = self.repo.mq.series[:]
         except AttributeError:
@@ -814,9 +838,10 @@ class RepoWidget(QWidget):
         if not cw.canswitch():
             return
         ctx = self.repo.changectx(rev)
-        if rev is None or ('mq' in self.repo.extensions() and 'qtip' in ctx.tags()):
-            # Clicking on working copy switches to commit tab
-            self._refreshCommitTabIfNeeded()
+        if rev is None or ('mq' in self.repo.extensions() and 'qtip' in ctx.tags()
+                           and self.repo['.'].rev() == rev):
+            # Clicking on working copy or on the topmost applied patch
+            # (_if_ it is also the working copy parent) switches to the commit tab
             tw.setCurrentIndex(self.commitTabIndex)
         else:
             # Clicking on a normal revision switches from commit tab
@@ -948,10 +973,12 @@ class RepoWidget(QWidget):
         self.revDetailsWidget.reload()
         self.filterbar.refresh()
 
+    @pyqtSlot()
     def repositoryDestroyed(self):
         'Repository has detected itself to be deleted'
         self.closeSelfSignal.emit(self)
 
+    @pyqtSlot()
     def repositoryChanged(self):
         'Repository has detected a changelog / dirstate change'
         if self.isVisible():
@@ -969,6 +996,7 @@ class RepoWidget(QWidget):
         # Update the repo registry entries related to the current repo
         self.repoChanged.emit(hglib.tounicode(self.repo.root))
 
+    @pyqtSlot()
     def configChanged(self):
         'Repository is reporting its config files have changed'
         self.repomodel.invalidate()
@@ -992,9 +1020,8 @@ class RepoWidget(QWidget):
             self.taskTabsWidget.tabBar().hide()
 
     @pyqtSlot(QString, bool)
-    def setBranch(self, branch, allparents=True):
+    def setBranch(self, branch, allparents):
         'Change the branch filter'
-        self.ubranch = branch
         self.repomodel.setBranch(branch=branch, allparents=allparents)
         self.titleChanged.emit(self.title())
         if self.revset:
@@ -1069,14 +1096,13 @@ class RepoWidget(QWidget):
         self.syncDemand.get().pull()
     def outgoing(self):
         self.syncDemand.get().outgoing()
-        self.outgoingMode = True
-    def push(self, confirm=True):
+    def push(self, confirm=True, **kwargs):
         """Call sync push.
 
         If confirm is False, the user will not be prompted for
         confirmation. If confirm is True, the prompt might be used.
         """
-        self.syncDemand.get().push(confirm)
+        self.syncDemand.get().push(confirm, **kwargs)
         self.outgoingMode = False
 
     @pyqtSlot()
@@ -1098,6 +1124,7 @@ class RepoWidget(QWidget):
         if len(selection) == 0:
             return
 
+        self.menuselection = selection
         if self.bundle:
             if len(selection) == 1:
                 self.bundlemenu.exec_(point)
@@ -1107,7 +1134,6 @@ class RepoWidget(QWidget):
                 self.outgoingcmenu.exec_(point)
                 return
 
-        self.menuselection = selection
         allunapp = False
         if 'mq' in self.repo.extensions():
             for rev in selection:
@@ -1196,100 +1222,102 @@ class RepoWidget(QWidget):
             if ext and ext not in exs:
                 return
             if desc is None:
-                menu.addSeparator()
-                return
+                return menu.addSeparator()
             act = QAction(desc, self)
-            act.triggered.connect(cb)
+            if cb:
+                act.triggered.connect(cb)
             if icon:
                 act.setIcon(qtlib.getmenuicon(icon))
             act.enableFunc = func
             menu.addAction(act)
             items.append(act)
+            return act
         menu = QMenu(self)
         if mode == 'outgoing':
-            submenu = menu.addMenu(_('Push'))
-            entry(submenu, None, isrev, _('Push all'), 'hg-push',
+            submenu = menu.addMenu(_('Pus&h'))
+            entry(submenu, None, isrev, _('Push &All'), 'hg-push',
+                  self.pushAll)
+            entry(submenu, None, isrev, _('Push to &Here'), '',
                   self.pushToRevision)
-            entry(submenu, None, isrev, _('Push to here'), '',
-                  self.pushToRevision)
-            entry(submenu, None, isrev, _('Push selected branch'), '',
+            entry(submenu, None, isrev, _('Push Selected &Branch'), '',
                   self.pushBranch)
             entry(menu)
-        entry(menu, None, isrev, _('Update...'), 'hg-update',
+        entry(menu, None, isrev, _('&Update...'), 'hg-update',
               self.updateToRevision)
         entry(menu)
-        entry(menu, None, isctx, _('Visual diff...'), 'visualdiff',
+        entry(menu, None, isctx, _('&Diff to Parent'), 'visualdiff',
               self.visualDiffRevision)
-        entry(menu, None, isrev, _('Diff to local...'), 'ldiff',
+        entry(menu, None, isrev, _('Diff to &Local'), 'ldiff',
               self.visualDiffToLocal)
-        entry(menu, None, isrev, _('Similar revisions...'), 'view-filter',
+        entry(menu, None, isctx, _('Bro&wse at Revision'), 'hg-annotate',
+              self.manifestRevision)
+        entry(menu, None, isrev, _('&Similar Revisions...'), 'view-filter',
               self.matchRevision)
         entry(menu)
-        entry(menu, None, fixed, _('Merge with local...'), 'hg-merge',
+        entry(menu, None, fixed, _('&Merge with Local...'), 'hg-merge',
               self.mergeWithRevision)
         entry(menu)
-        entry(menu, None, fixed, _('Tag...'), 'hg-tag',
+        entry(menu, None, fixed, _('&Tag...'), 'hg-tag',
               self.tagToRevision)
-        entry(menu, None, fixed, _('Bookmark...'), 'hg-bookmarks',
+        entry(menu, None, fixed, _('Boo&kmark...'), 'hg-bookmarks',
               self.bookmarkRevision)
         entry(menu)
-        entry(menu, None, fixed, _('Backout...'), 'hg-revert',
+        entry(menu, None, fixed, _('&Backout...'), 'hg-revert',
               self.backoutToRevision)
         entry(menu)
 
-        entry(menu, None, isrev, _('Copy hash'), 'copy-hash',
+        entry(menu, None, isrev, _('Copy &Hash'), 'copy-hash',
               self.copyHash)
         entry(menu)
 
-        submenu = menu.addMenu(_('Export'))
-        entry(submenu, None, isrev, _('Export patch...'), 'hg-export',
+        submenu = menu.addMenu(_('E&xport'))
+        entry(submenu, None, isrev, _('E&xport Patch...'), 'hg-export',
               self.exportRevisions)
-        entry(submenu, None, isrev, _('Email patch...'), 'mail-forward',
+        entry(submenu, None, isrev, _('&Email Patch...'), 'mail-forward',
               self.emailRevision)
-        entry(submenu, None, isrev, _('Archive...'), 'hg-archive',
+        entry(submenu, None, isrev, _('&Archive...'), 'hg-archive',
               self.archiveRevision)
-        entry(submenu, None, isrev, _('Bundle rev to tip...'), 'menurelocate',
-              self.bundleRevisions)
-        entry(submenu, None, isctx, _('Copy patch'), 'copy-patch',
+        entry(submenu, None, isrev, _('&Bundle Rev and Descendants...'),
+              'menurelocate', self.bundleRevisions)
+        entry(submenu, None, isctx, _('&Copy Patch'), 'copy-patch',
               self.copyPatch)
         entry(menu)
 
-        # hg >= 2.1
-        if hasattr(commands, 'phase'):
-            submenu = menu.addMenu(_('Change phase to'))
-            for pnum, pname in enumerate(phases.phasenames):
-                entry(submenu, None, isrev, pname, None,
-                      functools.partial(self.changePhase, pnum))
-            entry(menu)
+        submenu = menu.addMenu(_('Change &Phase to'))
+        submenu.triggered.connect(self._changePhaseByMenu)
+        for pnum, pname in enumerate(phases.phasenames):
+            entry(submenu, None, isrev, pname).setData(pnum)
+        entry(menu)
 
-        entry(menu, 'transplant', fixed, _('Transplant to local'), 'hg-transplant',
-              self.transplantRevisions)
+        entry(menu, None, fixed, _('&Graft to Local...'), 'hg-transplant',
+              self.graftRevisions)
 
         if 'mq' in exs or 'rebase' in exs:
-            submenu = menu.addMenu(_('Modify history'))
-            entry(submenu, 'mq', qgoto, _('Unapply patch (QGoto parent)'), 'hg-qgoto',
+            submenu = menu.addMenu(_('Modi&fy History'))
+            entry(submenu, 'mq', qgoto, _('&Unapply Patch (qgoto parent)'), 'hg-qgoto',
                   self.qgotoParentRevision)
-            entry(submenu, 'mq', fixed, _('Import to MQ'), 'qimport',
+            entry(submenu, 'mq', fixed, _('Import to &MQ'), 'qimport',
                   self.qimportRevision)
-            entry(submenu, 'mq', applied, _('Finish patch'), 'qfinish',
+            entry(submenu, 'mq', applied, _('&Finish Patch'), 'qfinish',
                   self.qfinishRevision)
-            entry(submenu, 'mq', applied, _('Rename patch...'), None,
+            entry(submenu, 'mq', applied, _('Re&name Patch...'), None,
                   self.qrename)
             entry(submenu, 'mq')
-            entry(submenu, 'rebase', fixed, _('Rebase...'), 'hg-rebase',
+            entry(submenu, 'rebase', fixed, _('&Rebase...'), 'hg-rebase',
                   self.rebaseRevision)
             entry(submenu, 'rebase')
-            entry(submenu, 'mq', fixed, _('Strip...'), 'menudelete',
+            entry(submenu, 'mq', fixed, _('&Strip...'), 'menudelete',
                   self.stripRevision)
 
-        entry(menu, 'reviewboard', fixed, _('Post to Review Board...'), 'reviewboard',
+        entry(menu, 'reviewboard', isrev, _('Post to Re&view Board...'), 'reviewboard',
               self.sendToReviewBoard)
 
-        entry(menu, 'rupdate', fixed, _('Remote Update...'), 'hg-update',
+        entry(menu, 'rupdate', fixed, _('&Remote Update...'), 'hg-update',
               self.rupdate)
 
         def _setupCustomSubmenu(menu):
-            tools, toolnames = hglib.tortoisehgtools(self.repo.ui, 'repowidget')
+            tools, toollist = hglib.tortoisehgtools(self.repo.ui,
+                selectedlocation='workbench.revdetails.custom-menu')
             if not tools:
                 return
 
@@ -1301,10 +1329,13 @@ class RepoWidget(QWidget):
 
             entry(menu)
             submenu = menu.addMenu(_('Custom Tools'))
-            for name in toolnames:
-                info = tools[name]
-                location = info.get('location', '').replace(' ', '').split(',')
-                if 'repowidget' not in location:
+            submenu.triggered.connect(self._runCustomCommandByMenu)
+            for name in toollist:
+                if name == '|':
+                    entry(submenu)
+                    continue
+                info = tools.get(name, None)
+                if info is None:
                     continue
                 command = info.get('command', None)
                 if not command:
@@ -1317,9 +1348,8 @@ class RepoWidget(QWidget):
                     enable = enablefuncs[enable]
                 else:
                     continue
-                menufunc = functools.partial(self.runCustomCommand, command,
-                    showoutput)
-                entry(submenu, None, enable, label, icon, menufunc)
+                a = entry(submenu, None, enable, label, icon)
+                a.setData((command, showoutput))
 
         _setupCustomSubmenu(menu)
 
@@ -1330,6 +1360,13 @@ class RepoWidget(QWidget):
             self.singlecmenu = menu
             self.singlecmenuitems = items
 
+    def _gotoAncestor(self):
+        ancestor = self.repo[self.menuselection[0]]
+        for rev in self.menuselection[1:]:
+            ctx = self.repo[rev]
+            ancestor = ancestor.ancestor(ctx)
+        self.goto(ancestor.rev())
+
     def generatePairMenu(self):
         def dagrange():
             revA, revB = self.menuselection
@@ -1407,10 +1444,6 @@ class RepoWidget(QWidget):
             dlg = compress.CompressDialog(self.repo, revs, self)
             dlg.finished.connect(dlg.deleteLater)
             dlg.exec_()
-        def gotoAncestor():
-            ctxa = self.repo[self.menuselection[0]]
-            ctxb = self.repo[self.menuselection[1]]
-            self.goto(ctxa.ancestor(ctxb).rev())
 
         menu = QMenu(self)
         for name, cb, icon in (
@@ -1428,8 +1461,10 @@ class RepoWidget(QWidget):
                 (_('Bisect - Bad, Good...'), bisectReverse, 'hg-bisect-bad-good'),
                 (_('Compress History...'), compressDlg, 'hg-compress'),
                 (None, None, None),
-                (_('Goto common ancestor'), gotoAncestor, 'hg-merge'),
-                (_('Similar revisions...'), self.matchRevision, 'view-filter')
+                (_('Goto common ancestor'), self._gotoAncestor, 'hg-merge'),
+                (_('Similar revisions...'), self.matchRevision, 'view-filter'),
+                (None, None, None),
+                (_('Graft Selected to local...'), self.graftRevisions, 'hg-transplant'),
                 ):
             if name is None:
                 menu.addSeparator()
@@ -1440,13 +1475,6 @@ class RepoWidget(QWidget):
             a.triggered.connect(cb)
             menu.addAction(a)
 
-        if 'transplant' in self.repo.extensions():
-            menu.addSeparator()
-            a = QAction(_('Transplant Selected to local'), self)
-            a.setIcon(qtlib.getmenuicon('hg-transplant'))
-            a.triggered.connect(self.transplantRevisions)
-            menu.addAction(a)
-
         if 'reviewboard' in self.repo.extensions():
             menu.addSeparator()
             a = QAction(_('Post Selected to Review Board...'), self)
@@ -1514,7 +1542,10 @@ class RepoWidget(QWidget):
                 (_('Export Selected...'), exportSel, 'hg-export'),
                 (_('Email Selected...'), emailSel, 'mail-forward'),
                 (None, None, None),
+                (_('Goto common ancestor'), self._gotoAncestor, 'hg-merge'),
                 (_('Similar revisions...'), self.matchRevision, 'view-filter'),
+                (None, None, None),
+                (_('Graft Selected to local...'), self.graftRevisions, 'hg-transplant'),
                 ):
             if name is None:
                 menu.addSeparator()
@@ -1525,13 +1556,6 @@ class RepoWidget(QWidget):
             a.triggered.connect(cb)
             menu.addAction(a)
 
-        if 'transplant' in self.repo.extensions():
-            menu.addSeparator()
-            a = QAction(_('Transplant Selected to local'), self)
-            a.setIcon(qtlib.getmenuicon('hg-transplant'))
-            a.triggered.connect(self.transplantRevisions)
-            menu.addAction(a)
-
         if 'reviewboard' in self.repo.extensions():
             a = QAction(_('Post Selected to Review Board...'), self)
             a.triggered.connect(self.sendToReviewBoard)
@@ -1678,7 +1702,22 @@ class RepoWidget(QWidget):
             dlg.deleteLater()
 
     def updateToRevision(self):
-        dlg = update.UpdateDialog(self.repo, self.rev, self)
+        ctx = self.repo[self.rev]
+        bookmarks = ctx.bookmarks()
+        if ctx in self.repo.parents():
+            # keep bookmark unchanged when updating to current rev
+            if self.repo._bookmarkcurrent in bookmarks:
+                rev = self.repo._bookmarkcurrent
+            else:
+                rev = self.rev
+        else:
+            # more common switching bookmark, rather than deselecting it
+            if bookmarks:
+                rev = bookmarks[0]
+            else:
+                rev = self.rev
+
+        dlg = update.UpdateDialog(self.repo, rev, self)
         dlg.output.connect(self.output)
         dlg.makeLogVisible.connect(self.makeLogVisible)
         dlg.progress.connect(self.progress)
@@ -1686,28 +1725,15 @@ class RepoWidget(QWidget):
         dlg.exec_()
 
     def matchRevision(self):
-        hasmatching = True
-        try:
-            # hg >= 2.2
-            from mercurial.revset import matching as matchingkeyword
-        except:
-            hasmatching = False
-        if hasmatching:
-            revlist = self.rev
-            if len(self.menuselection) > 1:
-                revlist = '|'.join([str(rev) for rev in self.menuselection])
-            dlg = matching.MatchDialog(self.repo, revlist, self)
-            if dlg.exec_():
-                self.setFilter(dlg.revsetexpression)
-        else:
-            # We cannot find similar revisions
-            # without the matching revset keyword
-            qtlib.WarningMsgBox(_('Incorrect Mercurial version'),
-                _('In order to use the "Find Similar revisions" '
-                'functionality, you must use a mercurial version above 2.1.'))
+        revlist = self.rev
+        if len(self.menuselection) > 1:
+            revlist = '|'.join([str(rev) for rev in self.menuselection])
+        dlg = matching.MatchDialog(self.repo, revlist, self)
+        if dlg.exec_():
+            self.setFilter(dlg.revsetexpression)
 
     def pushAll(self):
-        self.syncDemand.forward('push', True)
+        self.syncDemand.forward('push', False, pushall=True)
 
     def pushToRevision(self):
         # Do not ask for confirmation
@@ -1718,7 +1744,16 @@ class RepoWidget(QWidget):
         self.syncDemand.forward('push', False,
             branch=self.repo[self.rev].branch())
 
+    def manifestRevision(self):
+        run.manifest(self.repo.ui, repo=self.repo, rev=self.rev)
+
     def mergeWithRevision(self):
+        pctx = self.repo['.']
+        octx = self.repo[self.rev]
+        if pctx == octx:
+            QMessageBox.warning(self, _('Unable to merge'),
+                _('You cannot merge a revision with itself'))
+            return
         dlg = merge.MergeDialog(self.rev, self.repo, self)
         dlg.exec_()
         dlg.deleteLater()
@@ -1739,11 +1774,17 @@ class RepoWidget(QWidget):
         dlg.finished.connect(dlg.deleteLater)
         dlg.exec_()
 
-    def transplantRevisions(self):
-        cmdline = ['transplant', '--repository', self.repo.root]
-        for rev in self.repoview.selectedRevisions():
-            cmdline.append(str(rev))
-        self.runCommand(cmdline)
+    def graftRevisions(self):
+        """Graft selected revision on top of working directory parent"""
+        revlist = []
+        for rev in sorted(self.repoview.selectedRevisions()):
+            revlist.append(str(rev))
+        if not revlist:
+            revlist = [self.rev]
+        dlg = graft.GraftDialog(self.repo, self, source=revlist)
+        dlg.finished.connect(dlg.deleteLater)
+        if dlg.valid:
+            dlg.exec_()
 
     def backoutToRevision(self):
         dlg = backout.BackoutDialog(self.rev, self.repo, self)
@@ -1788,7 +1829,7 @@ class RepoWidget(QWidget):
             base = self.rev
         data = dict(name=os.path.basename(root), base=base)
         if tip is None:
-            filename = '%(name)s_%(base)s_to_tip.hg' % data
+            filename = '%(name)s_%(base)s_and_descendants.hg' % data
         else:
             data.update(rev=tip)
             filename = '%(name)s_%(base)s_to_%(rev)s.hg' % data
@@ -1816,10 +1857,7 @@ class RepoWidget(QWidget):
         _ui.pushbuffer()
         try:
             if self.rev and len(self.menuselection) == 1:
-                class Writable(object):
-                    def write(self, *args, **opts): _ui.write(*args, **opts)
-                    def close(self): pass
-                commands.export(_ui, self.repo, self.rev, output=Writable())
+                commands.export(_ui, self.repo, self.rev, output='')
             else:
                 revs = self.rev and self.menuselection or None
                 commands.diff(_ui, self.repo, rev=revs)
@@ -1836,7 +1874,10 @@ class RepoWidget(QWidget):
         if returnval:
             return output
         else:
-            QApplication.clipboard().setText(hglib.tounicode(output))
+            mdata = QMimeData()
+            mdata.setData('text/x-diff', output)  # for lossless import
+            mdata.setText(hglib.tounicode(output))
+            QApplication.clipboard().setMimeData(mdata)
 
     def copyHash(self):
         clip = QApplication.clipboard()
@@ -1861,8 +1902,8 @@ class RepoWidget(QWidget):
                     'Moving the phase backwards is not recommended.\n'
                     'For example, it may result in having multiple heads\nif you '
                     'modify a revision that you have already pushed\nto a server.\n\n'
-                    'Please be careful!' % (self.rev, phases.phasenames[currentphase], phasestr, phasestr,
-                                            phases.phasenames[currentphase]))
+                    'Please be careful!') % (self.rev, phases.phasenames[currentphase], phasestr, phasestr,
+                                            phases.phasenames[currentphase])
             labels = ((QMessageBox.Yes, _('&Force')),
                       (QMessageBox.No, _('&Cancel')))
             if not qtlib.QuestionMsgBox(title, main, text,
@@ -1872,6 +1913,11 @@ class RepoWidget(QWidget):
         self.runCommand(cmdlines)
         self.reload()
 
+    @pyqtSlot(QAction)
+    def _changePhaseByMenu(self, action):
+        phasenum, _ok = action.data().toInt()
+        self.changePhase(phasenum)
+
     def rebaseRevision(self):
         """Rebase selected revision on top of working directory parent"""
         opts = {'source' : self.rev, 'dest': self.repo['.'].rev()}
@@ -2012,7 +2058,7 @@ class RepoWidget(QWidget):
 
     def onCommandFinished(self, ret):
         self.repo.decrementBusyCount()
-        shlib.shell_notify(self.repo.root)
+        shlib.shell_notify([self.repo.root])
 
 
     def runCustomCommand(self, command, showoutput=False):
@@ -2029,6 +2075,7 @@ class RepoWidget(QWidget):
         # 2. Expand internal workbench variables
         vars = {
             'ROOT': self.repo.root,
+            'REVID': str(self.repo[self.rev]),
             'REV': self.rev
         }
         for var in vars:
@@ -2047,7 +2094,21 @@ class RepoWidget(QWidget):
             return self.runCommand(cmd)
 
         # Otherwise, run the selected command in the brackground
-        return subprocess.Popen(command, cwd=self.repo.root)
+        try:
+            res = subprocess.Popen(command, cwd=self.repo.root)
+        except OSError, ex:
+            res = 1
+            qtlib.ErrorMsgBox(_('Failed to execute custom command'),
+                _('The command "%s" could not be executed.') % hglib.tounicode(command),
+                _('The following error message was returned:\n\n"%s"\n\n'
+                'Please check that the command path is valid and '
+                'that it is a valid application') % hglib.tounicode(ex.strerror))
+        return res
+
+    @pyqtSlot(QAction)
+    def _runCustomCommandByMenu(self, action):
+        command, showoutput = action.data().toPyObject()
+        self.runCustomCommand(command, showoutput)
 
     def runCommand(self, *cmdlines):
         if self.runner.core.running():
@@ -2057,8 +2118,12 @@ class RepoWidget(QWidget):
         self.repo.incrementBusyCount()
         self.runner.run(*cmdlines)
 
+    @pyqtSlot()
     def _refreshCommitTabIfNeeded(self):
         """Refresh the Commit tab if the user settings require it"""
+        if self.taskTabsWidget.currentIndex() != self.commitTabIndex:
+            return
+
         refreshwd = self.repo.ui.config('tortoisehg', 'refreshwdstatus', 'auto')
         # Valid refreshwd values are 'auto', 'always' and 'alwayslocal'
         if refreshwd != 'auto':
