@@ -7,7 +7,7 @@
 # GNU General Public License version 2, incorporated herein by reference.
 
 shortlicense = '''
-Copyright (C) 2008-2012 Steve Borho <steve@borho.org> and others.
+Copyright (C) 2008-2013 Steve Borho <steve@borho.org> and others.
 This is free software; see the source for copying conditions.  There is NO
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 '''
@@ -29,8 +29,7 @@ from mercurial import util, fancyopts, cmdutil, extensions, error, scmutil
 from tortoisehg.hgqt.i18n import agettext as _
 from tortoisehg.util import hglib, paths, i18n
 from tortoisehg.util import version as thgversion
-from tortoisehg.hgqt import qtlib
-from tortoisehg.hgqt.bugreport import run as bugrun
+from tortoisehg.hgqt import bugreport, qtlib
 
 try:
     from tortoisehg.util.config import nofork as config_nofork
@@ -64,8 +63,8 @@ def dispatch(args):
         errstring = _('Error string "%(arg0)s" at %(arg1)s<br>Please '
                       '<a href="#edit:%(arg1)s">edit</a> your config')
         main = QApplication(sys.argv)
-        dlg = ExceptionMsgBox(hglib.tounicode(str(e)), errstring, opts,
-                              parent=None)
+        dlg = ExceptionMsgBox(hglib.tounicode(str(e)),
+                              hglib.tounicode(errstring), opts, parent=None)
         dlg.exec_()
     except SystemExit:
         pass
@@ -76,8 +75,10 @@ def dispatch(args):
         opts = {}
         opts['cmd'] = ' '.join(sys.argv[1:])
         opts['error'] = traceback.format_exc()
-        opts['nofork'] = True
-        return qtrun(bugrun, u, **opts)
+        main = QApplication(sys.argv)
+        dlg = bugreport.BugReport(opts)
+        dlg.exec_()
+        return -1
     except KeyboardInterrupt:
         print _('\nCaught keyboard interrupt, aborting.\n')
 
@@ -397,22 +398,20 @@ class GarbageCollector(QObject):
         #gc.set_debug(gc.DEBUG_SAVEALL)
 
     def check(self):
-        #return self.debug_cycles()
         l0, l1, l2 = gc.get_count()
-        if self.debug:
-            print 'gc_check called:', l0, l1, l2
         if l0 > self.threshold[0]:
             num = gc.collect(0)
             if self.debug:
-                print 'collecting gen 0, found:', num, 'unreachable'
+                print 'GarbageCollector.check:', l0, l1, l2
+                print 'collected gen 0, found', num, 'unreachable'
             if l1 > self.threshold[1]:
                 num = gc.collect(1)
                 if self.debug:
-                    print 'collecting gen 1, found:', num, 'unreachable'
+                    print 'collected gen 1, found', num, 'unreachable'
                 if l2 > self.threshold[2]:
                     num = gc.collect(2)
                     if self.debug:
-                        print 'collecting gen 2, found:', num, 'unreachable'
+                        print 'collected gen 2, found', num, 'unreachable'
 
     def debug_cycles(self):
         gc.collect()
@@ -504,8 +503,8 @@ class _QtRunner(QObject):
                                   hglib.tounicode(errstr), opts,
                                   parent=self._mainapp.activeWindow())
         elif etype is KeyboardInterrupt:
-            if qtlib.QuestionMsgBox(_('Keyboard interrupt'),
-                    _('Close this application?')):
+            if qtlib.QuestionMsgBox(hglib.tounicode(_('Keyboard interrupt')),
+                    hglib.tounicode(_('Close this application?'))):
                 QApplication.quit()
             else:
                 self.errors = []
@@ -528,8 +527,8 @@ class _QtRunner(QObject):
         QSettings.setDefaultFormat(QSettings.IniFormat)
 
         self._mainapp = QApplication(sys.argv)
-        self._gc = GarbageCollector(self, self.debug)
         try:
+            self._gc = GarbageCollector(self, self.debug)
             # default org is used by QSettings
             self._mainapp.setApplicationName('TortoiseHgQt')
             self._mainapp.setOrganizationName('TortoiseHg')
@@ -550,13 +549,15 @@ class _QtRunner(QObject):
                     from tortoisehg.hgqt import thgrepo
                     thgrepo.repository(ui, opts['repository'])
                 except error.RepoError, e:
-                    qtlib.WarningMsgBox(_('Repository Error'),
+                    qtlib.WarningMsgBox(hglib.tounicode(_('Repository Error')),
                                         hglib.tounicode(str(e)))
                     return
             dlg = dlgfunc(ui, *args, **opts)
             if dlg:
                 dlg.show()
                 dlg.raise_()
+            else:
+                return -1
         except:
             # Exception before starting eventloop needs to be postponed;
             # otherwise it will be ignored silently.
@@ -639,6 +640,11 @@ def email(ui, *pats, **opts):
     from tortoisehg.hgqt.hgemail import run
     return qtrun(run, ui, *pats, **opts)
 
+def graft(ui, *revs, **opts):
+    """graft dialog"""
+    from tortoisehg.hgqt.graft import run
+    return qtrun(run, ui, *revs, **opts)
+
 def resolve(ui, *pats, **opts):
     """resolve dialog"""
     from tortoisehg.hgqt.resolve import run
@@ -1076,7 +1082,7 @@ table = {
              '(only a repository)')),
           ('u', 'updaterev', '',
            _('revision, tag or branch to check out')),
-          ('r', 'rev', [], _('include the specified changeset')),
+          ('r', 'rev', '', _('include the specified changeset')),
           ('b', 'branch', [],
            _('clone only the specified branch')),
           ('', 'pull', None, _('use pull protocol to copy metadata')),
@@ -1089,6 +1095,9 @@ table = {
         _('thg commit [OPTIONS] [FILE]...')),
     "drag_move": (drag_move, [], _('thg drag_move SOURCE... DEST')),
     "drag_copy": (drag_copy, [], _('thg drag_copy SOURCE... DEST')),
+    "graft": (graft,
+        [('r', 'rev', [], _('revisions to graft'))],
+        _('thg graft [-r] REV...')),
     "^grep|search": (grep,
         [('i', 'ignorecase', False, _('ignore case during search')),],
         _('thg grep')),
@@ -1104,7 +1113,7 @@ table = {
          _('thg email [REVS]')),
     "^log|history|explorer|workbench":
         (log,
-         [('l', 'limit', '', _('limit number of changes displayed'))],
+         [('l', 'limit', '', _('(DEPRECATED)'))],
          _('thg log [OPTIONS] [FILE]')),
     "manifest":
         (manifest,
@@ -1142,13 +1151,12 @@ table = {
     "^rebase": (rebase,
         [('', 'keep', False, _('keep original changesets')),
          ('', 'keepbranches', False, _('keep original branch names')),
-         ('', 'detach', False, _('force detaching of source from its original '
-                                'branch')),
+         ('', 'detach', False, _('(DEPRECATED)')),
          ('s', 'source', '',
           _('rebase from the specified changeset')),
          ('d', 'dest', '',
           _('rebase onto the specified changeset'))],
-        _('thg rebase -s REV -d REV [--keep] [--detach]')),
+        _('thg rebase -s REV -d REV [--keep]')),
     "^tag":
         (tag,
          [('f', 'force', None, _('replace existing tag')),
