@@ -62,6 +62,10 @@ class StatusWidget(QWidget):
         self.pctx = None
         self.savechecks = True
         self.refthread = None
+        self.refreshWctxLater = QTimer(self, interval=10, singleShot=True)
+        self.refreshWctxLater.timeout.connect(self.refreshWctx)
+        self.partials = {}
+        self.manualCheckAllUpdate = False
 
         # determine the user configured status colors
         # (in the future, we could support full rich-text tags)
@@ -185,6 +189,8 @@ class StatusWidget(QWidget):
         self.fileview.linkActivated.connect(self.linkActivated)
         self.fileview.fileDisplayed.connect(self.fileDisplayed)
         self.fileview.shelveToolExited.connect(self.refreshWctx)
+        self.fileview.newChunkList.connect(self.updatePartials)
+        self.fileview.chunkSelectionChanged.connect(self.chunkSelectionChanged)
         self.fileview.setContext(self.repo[None])
         self.fileview.setMinimumSize(QSize(16, 16))
         vbox.addWidget(self.fileview, 1)
@@ -193,6 +199,8 @@ class StatusWidget(QWidget):
         self.diffvbox = vbox
 
     def checkAllNone(self):
+        if self.manualCheckAllUpdate:
+            return
         state = self.checkAllNoneBtn.checkState()
         if state == Qt.Checked:
             self.checkAll()
@@ -218,6 +226,53 @@ class StatusWidget(QWidget):
         self.fileview.saveSettings(qs, prefix+'/fileview')
         qs.setValue(prefix+'/state', self.split.saveState())
 
+    @pyqtSlot(QString, object)
+    def updatePartials(self, wfile, changes):
+        # remove files from the partials dictionary if they are not partial
+        # selections, in order to simplify refresh.
+        dels = []
+        for file, oldchanges in self.partials.iteritems():
+            if oldchanges.excludecount == 0:
+                self.tv.model().checked[file] = True
+                dels.append(file)
+            elif oldchanges.excludecount == len(oldchanges.hunks):
+                self.tv.model().checked[file] = False
+                dels.append(file)
+        for file in dels:
+            del self.partials[file]
+
+        if changes is None:
+            if wfile in self.partials:
+                del self.partials[wfile]
+            return
+
+        wfile = hglib.fromunicode(wfile)
+        if wfile in self.partials:
+            # merge selection state from old hunk list to new hunk list
+            oldhunks = self.partials[wfile].hunks
+            oldstates = dict([(c.fromline, c.excluded) for c in oldhunks])
+            for chunk in changes.hunks:
+                if chunk.fromline in oldstates:
+                    self.fileview.updateChunk(chunk, oldstates[chunk.fromline])
+        else:
+            # the file was not in the partials dictionary, so it is either
+            # checked (all changes enabled) or unchecked (all changes
+            # excluded).
+            if wfile not in self.getChecked():
+                for chunk in changes.hunks:
+                    self.fileview.updateChunk(chunk, True)
+        self.chunkSelectionChanged()
+        self.partials[wfile] = changes
+
+    @pyqtSlot()
+    def chunkSelectionChanged(self):
+        'checkbox state has changed via chunk selection'
+        # inform filelist view that the file selection state may have changed
+        model = self.tv.model()
+        if model:
+            model.layoutChanged.emit()
+            model.checkCountChanged.emit()
+
     @pyqtSlot(QPoint, object)
     def onMenuRequest(self, point, selected):
         menu = self.actions.makeMenu(selected)
@@ -226,14 +281,16 @@ class StatusWidget(QWidget):
 
     def setPatchContext(self, pctx):
         if pctx != self.pctx:
+            # clear out the current checked state on next refreshWctx()
             self.savechecks = False
-        else:
-            self.savechecks = True
         self.pctx = pctx
 
+    @pyqtSlot()
     def refreshWctx(self, synchronous=False):
         if self.refthread:
+            self.refreshWctxLater.start()
             return
+        self.refreshWctxLater.stop()
         self.fileview.clearDisplay()
 
         # store selected paths or current path
@@ -262,6 +319,7 @@ class StatusWidget(QWidget):
         if synchronous:
             self.reloadComplete()
 
+    @pyqtSlot()
     def reloadComplete(self):
         self.refthread.wait()
         if self.checkable:
@@ -271,6 +329,12 @@ class StatusWidget(QWidget):
         if self.refthread.wctx is not None:
             self.updateModel(self.refthread.wctx, self.refthread.patchecked)
         self.refthread = None
+        if len(self.repo.parents()) > 1:
+            # nuke partial selections if wctx has a merge in-progress
+            self.partials = {}
+        match = self.le.text()
+        if match:
+            self.setFilter(match)
 
     def canExit(self):
         return not self.refthread
@@ -290,9 +354,14 @@ class StatusWidget(QWidget):
                        checked, self, checkable=self.checkable,
                        defcheck=self.defcheck)
         if self.checkable:
-            tm.checkToggled.connect(self.updateCheckCount)
+            tm.checkToggled.connect(self.checkToggled)
+            tm.checkCountChanged.connect(self.updateCheckCount)
+        self.savechecks = True
 
+        oldtm = self.tv.model()
         self.tv.setModel(tm)
+        if oldtm:
+            oldtm.deleteLater()
         self.tv.setSortingEnabled(True)
         self.tv.setColumnHidden(COL_PATH, bool(wctx.p2()) or not self.checkable)
         self.tv.setColumnHidden(COL_MERGE_STATE, not tm.anyMerge())
@@ -340,7 +409,7 @@ class StatusWidget(QWidget):
             self.actions.allactions[0].trigger()
         elif status == 'S':
             self.linkActivated.emit(
-                u'subrepo:' + hglib.tounicode(self.repo.wjoin(path)))
+                u'repo:' + hglib.tounicode(self.repo.wjoin(path)))
         elif status in 'C?':
             qtlib.editfiles(self.repo, [path])
 
@@ -349,8 +418,11 @@ class StatusWidget(QWidget):
         model = self.tv.model()
         if model:
             model.setFilter(match)
+            self.tv.enablefilterpalette(bool(match))
 
+    @pyqtSlot()
     def updateCheckCount(self):
+        'user has toggled one or more checkboxes, update counts and checkall'
         model = self.tv.model()
         if model:
             model.checkCount = len(self.getChecked())
@@ -360,7 +432,20 @@ class StatusWidget(QWidget):
                 state = Qt.Checked
             else:
                 state = Qt.PartiallyChecked
+            self.manualCheckAllUpdate = True
             self.checkAllNoneBtn.setCheckState(state)
+            self.manualCheckAllUpdate = False
+
+    @pyqtSlot(QString, bool)
+    def checkToggled(self, wfile, checked):
+        'user has toggled a checkbox, update partial chunk selection status'
+        wfile = hglib.fromunicode(wfile)
+        if wfile in self.partials:
+            if wfile == self.fileview._filename:
+                for chunk in self.partials[wfile].hunks:
+                    self.fileview.updateChunk(chunk, not checked)
+            else:
+                del self.partials[wfile]
 
     def checkAll(self):
         model = self.tv.model()
@@ -377,13 +462,26 @@ class StatusWidget(QWidget):
         if model:
             checked = model.getChecked()
             if types is None:
-                return [f for f, v in checked.iteritems() if v]
+                files = []
+                for f, v in checked.iteritems():
+                    if f in self.partials:
+                        changes = self.partials[f]
+                        if changes.excludecount < len(changes.hunks):
+                            files.append(f)
+                    elif v:
+                        files.append(f)
+                return files
             else:
                 files = []
                 for row in model.getAllRows():
                     path, status, mst, upath, ext, sz = row
-                    if status in types and checked[path]:
-                        files.append(path)
+                    if status in types:
+                        if path in self.partials:
+                            changes = self.partials[path]
+                            if changes.excludecount < len(changes.hunks):
+                                files.append(path)
+                        elif checked[path]:
+                            files.append(path)
                 return files
         else:
             return []
@@ -497,6 +595,7 @@ class WctxFileTree(QTreeView):
         self.setContextMenuPolicy(Qt.CustomContextMenu)
         self.customContextMenuRequested.connect(self.menuRequested)
         self.setTextElideMode(Qt.ElideLeft)
+        self._paletteswitcher = qtlib.PaletteSwitcher(self)
 
     def scrollTo(self, index, hint=QAbstractItemView.EnsureVisible):
         # don't update horizontal position by selection change
@@ -550,12 +649,17 @@ class WctxFileTree(QTreeView):
         # Invalid selectionModel found
         return []
 
+    def enablefilterpalette(self, enable):
+        self._paletteswitcher.enablefilterpalette(enable)
+
 class WctxModel(QAbstractTableModel):
-    checkToggled = pyqtSignal()
+    checkCountChanged = pyqtSignal()
+    checkToggled = pyqtSignal(QString, bool)
 
     def __init__(self, wctx, ms, pctx, savechecks, opts, checked, parent,
                  checkable=True, defcheck='MAR!S'):
         QAbstractTableModel.__init__(self, parent)
+        self.partials = parent.partials
         self.checkCount = 0
         rows = []
         nchecked = {}
@@ -632,17 +736,19 @@ class WctxModel(QAbstractTableModel):
             return 0 # no child
         return len(self.rows)
 
-    def check(self, files, state=True):
+    def check(self, files, state):
         for f in files:
             self.checked[f] = state
+            self.checkToggled.emit(f, state)
         self.layoutChanged.emit()
-        self.checkToggled.emit()
-        
+        self.checkCountChanged.emit()
+
     def checkAll(self, state):
         for data in self.rows:
             self.checked[data[0]] = state
+            self.checkToggled.emit(data[3], state)
         self.layoutChanged.emit()
-        self.checkToggled.emit()
+        self.checkCountChanged.emit()
 
     def columnCount(self, parent):
         if parent.isValid():
@@ -656,7 +762,14 @@ class WctxModel(QAbstractTableModel):
         path, status, mst, upath, ext, sz = self.rows[index.row()]
         if index.column() == COL_PATH:
             if role == Qt.CheckStateRole and self.checkable:
-                # also Qt.PartiallyChecked
+                if path in self.partials:
+                    changes = self.partials[path]
+                    if changes.excludecount == 0:
+                        return Qt.Checked
+                    elif changes.excludecount == len(changes.hunks):
+                        return Qt.Unchecked
+                    else:
+                        return Qt.PartiallyChecked
                 if self.checked[path]:
                     return Qt.Checked
                 else:
@@ -721,9 +834,18 @@ class WctxModel(QAbstractTableModel):
         for index in indexes:
             assert index.isValid()
             fname = self.rows[index.row()][COL_PATH]
-            self.checked[fname] = not self.checked[fname]
+            uname = self.rows[index.row()][COL_PATH_DISPLAY]
+            if fname in self.partials:
+                checked = 0
+                changes = self.partials[fname]
+                if changes.excludecount < len(changes.hunks):
+                    checked = 1
+            else:
+                checked = self.checked[fname]
+            self.checked[fname] = not checked
+            self.checkToggled.emit(uname, self.checked[fname])
         self.layoutChanged.emit()
-        self.checkToggled.emit()
+        self.checkCountChanged.emit()
 
     def sort(self, col, order):
         self.layoutAboutToBeChanged.emit()
@@ -799,7 +921,8 @@ class WctxModel(QAbstractTableModel):
     def setFilter(self, match):
         'simple match in filename filter'
         self.layoutAboutToBeChanged.emit()
-        self.rows = [r for r in self.unfiltered if match in r[COL_PATH_DISPLAY]]
+        self.rows = [r for r in self.unfiltered
+                     if unicode(match) in r[COL_PATH_DISPLAY]]
         self.layoutChanged.emit()
         self.reset()
 
@@ -917,15 +1040,16 @@ class StatusDialog(QDialog):
         self.setWindowFlags(Qt.Window)
         self.loadSettings()
 
-        QShortcut(QKeySequence.Refresh, self, self.stwidget.refreshWctx)
+        qtlib.newshortcutsforstdkey(QKeySequence.Refresh, self,
+                                    self.stwidget.refreshWctx)
         QTimer.singleShot(0, self.stwidget.refreshWctx)
 
     def linkActivated(self, link):
         link = hglib.fromunicode(link)
-        if link.startswith('subrepo:'):
+        if link.startswith('repo:'):
             from tortoisehg.hgqt.run import qtrun
             from tortoisehg.hgqt import commit
-            qtrun(commit.run, self.stwidget.repo.ui, root=link[8:])
+            qtrun(commit.run, self.stwidget.repo.ui, root=link[len('repo:'):])
         if link.startswith('shelve:'):
             from tortoisehg.hgqt import shelve
             dlg = shelve.ShelveDialog(self.stwidget.repo, self)
