@@ -51,6 +51,8 @@ def repository(_ui=None, path='', create=False, bundle=None):
             _ui = uimod.ui()
         try:
             repo = hg.repository(_ui, path, create)
+            # get unfiltered repo in version safe manner
+            repo = getattr(repo, 'unfiltered', lambda: repo)()
             repo.__class__ = _extendrepo(repo)
             repo._pyqtobj = ThgRepoWrapper(repo)
             _repocache[path] = repo
@@ -178,8 +180,6 @@ class ThgRepoWrapper(QObject):
         watchedfiles = [self.repo.sjoin('00changelog.i')]
         watchedfiles.append(self.repo.sjoin('phaseroots'))
         watchedfiles.append(self.repo.join('localtags'))
-        watchedfiles.append(self.repo.join('bookmarks'))
-        watchedfiles.append(self.repo.join('bookmarks.current'))
         if hasattr(self.repo, 'mq'):
             watchedfiles.append(self.repo.mq.path)
             watchedfiles.append(self.repo.mq.join('series'))
@@ -272,12 +272,8 @@ class ThgRepoWrapper(QObject):
 _uiprops = '''_uifiles _uimtime postpull tabwidth maxdiff
               deadbranches _exts _thghiddentags displayname summarylen
               shortname mergetools namedbranches'''.split()
-
-# _bookmarkcurrent is a Mercurial property, we include it here to work
-# around a bug in hg-1.8.  It should be removed when we drop support for
-# Mercurial 1.8
 _thgrepoprops = '''_thgmqpatchnames thgmqunappliedpatches
-                   _branchheads _bookmarkcurrent'''.split()
+                   _branchheads'''.split()
 
 def _extendrepo(repo):
     class thgrepository(repo.__class__):
@@ -573,13 +569,13 @@ def _extendrepo(repo):
             if 'largefiles' in self.extensions() or 'kbfiles' in self.extensions():
                 path = _kbfregex.sub('', path)
             return path
-        
+
         def bfStandin(self, path):
             return '.kbf/' + path
 
         def lfStandin(self, path):
             return '.hglf/' + path
-        
+
     return thgrepository
 
 _maxchangectxclscache = 10
@@ -606,7 +602,11 @@ def _createchangectxcls(parentcls):
         def sub(self, path):
             srepo = super(thgchangectx, self).sub(path)
             if isinstance(srepo, subrepo.hgsubrepo):
-                srepo._repo.__class__ = _extendrepo(srepo._repo)
+                r = srepo._repo
+                # get unfiltered repo in version safe manner
+                r = getattr(r, 'unfiltered', lambda: r)()
+                r.__class__ = _extendrepo(r)
+                srepo._repo = r
             return srepo
 
         def thgtags(self):
@@ -670,7 +670,7 @@ def _createchangectxcls(parentcls):
                     summary += u' \u2026' # ellipsis ...
 
             return summary
-        
+
         def hasStandin(self, file):
             if 'largefiles' in self._repo.extensions():
                 if self._repo.lfStandin(file) in self.manifest():
@@ -682,10 +682,10 @@ def _createchangectxcls(parentcls):
 
         def isStandin(self, path):
             return self._repo.isStandin(path)
-        
+
         def removeStandin(self, path):
             return self._repo.removeStandin(path)
-        
+
         def findStandin(self, file):
             if 'largefiles' in self._repo.extensions():
                 if self._repo.lfStandin(file) in self.manifest():
