@@ -1,4 +1,4 @@
-# thread.py - A seprated thread to run Mercurial command
+# thread.py - A separate thread to run Mercurial commands
 #
 # Copyright 2009 Steve Borho <steve@borho.org>
 # Copyright 2010 Yuki KODAMA <endflow.net@gmail.com>
@@ -9,20 +9,27 @@
 import os
 import Queue
 import time
-import urllib2
+import urllib2, urllib
 import socket
 import errno
 
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
-from mercurial import util, error
+from mercurial import util, error, subrepo
 from mercurial import ui as uimod
 
 from tortoisehg.util import thread2, hglib
 from tortoisehg.hgqt.i18n import _, localgettext
 from tortoisehg.hgqt import qtlib
 
+try:
+    _SubrepoAbort = subrepo.SubrepoAbort
+except AttributeError:
+    # hg <= 2.4
+    class _SubrepoAbort(Exception):
+        pass
+
 local = localgettext()
 
 class DataWrapper(object):
@@ -48,17 +55,13 @@ class UiSignal(QObject):
         label = hglib.tounicode(opts.get('label', 'ui.error'))
         self.writeSignal.emit(msg, label)
 
-    def prompt(self, msg, choices, default):
+    def prompt(self, msg, default):
         try:
-            r = self._waitresponse(msg, False, choices, None)
+            r = self._waitresponse(msg, False, None, None)
             if r is None:
                 raise EOFError
             if not r:
                 return default
-            if choices:
-                # return char for Mercurial 1.3
-                choice = choices[r]
-                return choice[choice.index('&')+1].lower()
             return r
         except EOFError:
             raise util.Abort(local._('response expected'))
@@ -118,9 +121,9 @@ class QtUi(uimod.ui):
     def flush(self):
         pass
 
-    def prompt(self, msg, choices=None, default='y'):
+    def prompt(self, msg, default='y'):
         if not self.interactive(): return default
-        return self.sig.prompt(msg, choices, default)
+        return self.sig.prompt(msg, default)
 
     def promptchoice(self, msg, choices, default=0):
         if not self.interactive(): return default
@@ -272,6 +275,14 @@ class CmdThread(QThread):
                 ui.setconfig('defaults', k, '')
             self.ret = 255
             self.ret = hglib.dispatch(ui, self.cmdline) or 0
+        except _SubrepoAbort, e:
+            errormsg = str(e)
+            label = 'ui.error'
+            if e.subrepo:
+                label += ' subrepo=%s' % urllib.quote(e.subrepo)
+            ui.write_err(local._('abort: ') + errormsg + '\n', label=label)
+            if e.hint:
+                ui.write_err(local._('hint: ') + str(e.hint) + '\n', label=label)
         except util.Abort, e:
             ui.write_err(local._('abort: ') + str(e) + '\n')
             if e.hint:
