@@ -37,40 +37,41 @@ class WctxActions(QObject):
             parent.addAction(action)
             allactions.append(action)
 
-        make(_('&Visual Diff'), vdiff, frozenset('MAR!'), 'visualdiff', 'CTRL+D')
-        make(_('Copy patch'), copyPatch, frozenset('MAR!'), 'copy-patch')
-        make(_('Edit'), edit, frozenset('MACI?'), 'edit-file', 'SHIFT+CTRL+E')
-        make(_('Open'), openfile, frozenset('MACI?'), '', 'SHIFT+CTRL+O')
+        make(_('&Diff to Parent'), vdiff, frozenset('MAR!'), 'visualdiff', 'CTRL+D')
+        make(_('&Copy Patch'), copyPatch, frozenset('MAR!'), 'copy-patch')
+        make(_('&Edit'), edit, frozenset('MACI?'), 'edit-file', 'SHIFT+CTRL+E')
+        make(_('&Open'), openfile, frozenset('MACI?'), '', 'SHIFT+CTRL+L')
         allactions.append(None)
-        make(_('Open subrepository'), opensubrepo, frozenset('S'),
-            'thg-repository-open', 'ALT+CTRL+O')
-        make(_('Explore subrepository'), explore, frozenset('S'),
-            'system-file-manager', 'ALT+CTRL+E')
-        make(_('Open terminal in subrepository'), terminal, frozenset('S'),
-            'utilities-terminal', 'ALT+CTRL+T')
+        make(_('Open S&ubrepository'), opensubrepo, frozenset('S'),
+            'thg-repository-open')
+        make(_('E&xplore Subrepository'), explore, frozenset('S'),
+            'system-file-manager')
+        make(_('Open &Terminal'), terminal, frozenset('S'),
+            'utilities-terminal')
         allactions.append(None)
-        make(_('Copy path'), copyPath, frozenset('MARC?!IS'), '')
-        make(_('View missing'), viewmissing, frozenset('R!'))
+        make(_('Copy &Path'), copyPath, frozenset('MARC?!IS'), '')
+        make(_('&View Missing'), viewmissing, frozenset('R!'))
         allactions.append(None)
         make(_('&Revert...'), revert, frozenset('SMAR!'), 'hg-revert')
         make(_('&Add'), add, frozenset('R'), 'fileadd')
         allactions.append(None)
-        make(_('File History'), log, frozenset('MARC!'), 'hg-log')
+        make(_('File &History'), log, frozenset('MARC!'), 'hg-log')
         make(_('&Annotate'), annotate, frozenset('MARC!'), 'hg-annotate')
         allactions.append(None)
-        make(_('&Forget'), forget, frozenset('MAC!'), 'filedelete')
+        make(_('&Forget'), forget, frozenset('MC!'), 'filedelete')
         make(_('&Add'), add, frozenset('I?'), 'fileadd')
         if 'largefiles' in self.repo.extensions():
             make(_('Add &Largefiles...'), addlf, frozenset('I?'))
-        make(_('&Detect Renames...'), guessRename, frozenset('A?!'),
+        make(_('De&tect Renames...'), guessRename, frozenset('A?!'),
              'detect_rename')
         make(_('&Ignore...'), ignore, frozenset('?'), 'ignore')
-        make(_('Remove versioned'), remove, frozenset('C'), 'remove')
-        make(_('&Delete unversioned...'), delete, frozenset('?I'), 'hg-purge')
+        make(_('Re&move Versioned'), remove, frozenset('C'), 'remove')
+        make(_('&Delete Unversioned...'), delete, frozenset('?I'), 'hg-purge')
         allactions.append(None)
-        make(_('Mark unresolved'), unmark, frozenset('r'))
-        make(_('Mark resolved'), mark, frozenset('u'))
+        make(_('&Mark Unresolved'), unmark, frozenset('r'))
+        make(_('&Mark Resolved'), mark, frozenset('u'))
         if checkable:
+            # no &-shortcut because check/uncheck can be done by space key
             allactions.append(None)
             make(_('Check'), check, frozenset('MARC?!IS'), '')
             make(_('Uncheck'), uncheck, frozenset('MARC?!IS'), '')
@@ -116,12 +117,12 @@ class WctxActions(QObject):
             action.triggered.connect(self.runAction)
 
         if len(repo.parents()) > 1:
-            make(_('View other'), viewother, frozenset('MA'))
+            make(_('View O&ther'), viewother, frozenset('MA'))
 
         if len(selrows) == 1:
             menu.addSeparator()
             make(_('&Copy...'), copy, frozenset('MC'), 'edit-copy')
-            make(_('Rename...'), rename, frozenset('MC'), 'hg-rename')
+            make(_('Re&name...'), rename, frozenset('MC'), 'hg-rename')
 
         # Add 'was renamed from' actions for unknown files
         t, path = selrows[0]
@@ -138,9 +139,9 @@ class WctxActions(QObject):
 
         # Add restart merge actions for resolved files
         if alltypes & frozenset('u'):
-            f = make(_('Restart Merge...'), resolve, frozenset('u'))
+            f = make(_('Restart Mer&ge'), resolve, frozenset('u'))
             files = [f for t, f in selrows if 'u' in t]
-            rmenu = QMenu(_('Restart merge with'), self.parent())
+            rmenu = QMenu(_('Restart Merge &with'), self.parent())
             for tool in hglib.mergetools(repo.ui):
                 def mkaction(rtool):
                     a = rmenu.addAction(hglib.tounicode(rtool))
@@ -230,11 +231,11 @@ def openfile(parent, ui, repo, files):
 
 def opensubrepo(parent, ui, repo, files):
     for filename in files:
-        path = os.path.join(repo.root, files[0])
+        path = os.path.join(repo.root, filename)
         if os.path.isdir(path):
-            parent.linkActivated.emit(u'subrepo:'+hglib.tounicode(path))
+            parent.linkActivated.emit(u'repo:' + hglib.tounicode(path))
         else:
-            QMessageBox.warning(self,
+            QMessageBox.warning(parent,
                 _("Cannot open subrepository"),
                 _("The selected subrepository does not exist on the working directory"))
 
@@ -258,7 +259,7 @@ def viewother(parent, ui, repo, files):
     edit(parent, ui, repo, [os.path.join(base, f) for f in files])
 
 def revert(parent, ui, repo, files):
-    revertopts = {'date': None, 'rev': '.'}
+    revertopts = {'date': None, 'rev': '.', 'all': False}
 
     if len(repo.parents()) > 1:
         res = qtlib.CustomPrompt(
@@ -270,18 +271,27 @@ def revert(parent, ui, repo, files):
         elif res == 1:
             revertopts['rev'] = repo[None].p2().rev()
         else:
-            return
+            return False
         commands.revert(ui, repo, *files, **revertopts)
     else:
-        res = qtlib.CustomPrompt(
+        wctx = repo[None]
+        if [file for file in files if file in wctx.modified()]:
+            res = qtlib.CustomPrompt(
                 _('Confirm Revert'),
                 _('Revert local file changes?'), parent,
                 (_('&Revert with backup'), _('&Discard changes'),
                 _('Cancel')), 2, 2, files).run()
-        if res == 2:
-            return False
-        if res == 1:
-            revertopts['no_backup'] = True
+            if res == 2:
+                return False
+            if res == 1:
+                revertopts['no_backup'] = True
+        else:
+            res = qtlib.CustomPrompt(
+                    _('Confirm Revert'),
+                    _('Revert the following files?'),
+                    parent, (_('&Revert'), _('Cancel')), 1, 1, files).run()
+            if res == 1:
+                return False
         commands.revert(ui, repo, *files, **revertopts)
         return True
 
@@ -348,21 +358,17 @@ def delete(parent, ui, repo, files):
             _('Delete the following unversioned files?'),
             parent, (_('&Delete'), _('Cancel')), 1, 1, files).run()
     if res == 1:
-        return
+        return False
     for wfile in files:
         os.unlink(wfile)
     return True
 
 def copy(parent, ui, repo, files):
+    from tortoisehg.hgqt.rename import RenameDialog
     assert len(files) == 1
-    wfile = repo.wjoin(files[0])
-    fd = QFileDialog(parent)
-    fname = fd.getSaveFileName(parent, _('Copy file to'), wfile)
-    if not fname:
-        return
-    fname = hglib.fromunicode(fname)
-    wfiles = [wfile, fname]
-    commands.copy(ui, repo, *wfiles)
+    dlg = RenameDialog(ui, files, parent, iscopy=True)
+    dlg.finished.connect(dlg.deleteLater)
+    dlg.exec_()
     return True
 
 def rename(parent, ui, repo, files):
@@ -398,7 +404,7 @@ def resolve_with(tool, repo, files):
     return True
 
 def check(parent, ui, repo, files):
-    parent.tv.model().check(files)
+    parent.tv.model().check(files, True)
     return True
 
 def uncheck(parent, ui, repo, files):
